# Author: Daniel Heimans
# URL: http://code.google.com/p/sickrage
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
import uuid

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.enums import SearchFormat, SeriesProviderID
from sickrage.core.helpers import sanitize_scene_name, episode_num, try_int
from sickrage.core.tv.show.helpers import find_show
from sickrage.search_providers import TorrentProvider


class BTNProvider(TorrentProvider):
    def __init__(self):
        super(BTNProvider, self).__init__("BTN", 'https://broadcasthe.net', True)

        self._urls.update({
            'api': 'https://api.broadcasthe.net',
        })

        self.supports_absolute_numbering = True

        # custom settings
        self.custom_settings = {
            'api_key': '',
            'reject_m2ts': False,
            'minseed': 0,
            'minleech': 0
        }

        self.cache = TVCache(self, min_time=10)

    def _check_auth(self):
        if not self.custom_settings['api_key']:
            sickrage.app.log.warning("Missing/Invalid API key. Check your settings")
            return False
        return True

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        """
        Search a provider and parse the results.
        :param search_strings: A dict with {mode: search value}
        :param age: Not used
        :returns: A list of search results (structure)
        """
        results = []
        if not self._check_auth():
            return results

        # Search Params
        search_params = {
            'age': '<=10800',  # Results from the past 3 hours
        }

        for mode in search_strings:
            sickrage.app.log.debug('Search mode: {}'.format(mode))

            if mode != 'RSS':
                searches = self._search_params(series_id, series_provider_id, season, episode, mode)
            else:
                searches = [search_params]

            for search_params in searches:
                if mode != 'RSS':
                    sickrage.app.log.debug('Search string: {}'.format(search_params))

                response = self._api_call(search_params)
                if not response or response.get('results') == '0':
                    sickrage.app.log.debug('No data returned from provider')
                    continue

                results += self.parse(response.get('torrents', {}), mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results for items.
        :param data: The raw response from a search
        :param mode: The current mode used to search, e.g. RSS
        :return: A list of items found
        """
        results = []

        torrent_rows = data.values()

        for row in torrent_rows:
            title, download_url = self._process_title_and_url(row)
            if not all([title, download_url]):
                continue

            seeders = try_int(row.get('Seeders'))
            leechers = try_int(row.get('Leechers'))
            size = try_int(row.get('Size'), -1)

            results += [{
                'title': title,
                'link': download_url,
                'size': size,
                'seeders': seeders,
                'leechers': leechers
            }]

            sickrage.app.log.debug("Found result: {}".format(title))

        return results

    @staticmethod
    def _process_title_and_url(parsed_json, **kwargs):
        # The BTN API gives a lot of information in response,
        # however SickRage is built mostly around Scene or
        # release names, which is why we are using them here.

        if 'ReleaseName' in parsed_json and parsed_json['ReleaseName']:
            title = parsed_json['ReleaseName']

        else:
            # If we don't have a release name we need to get creative
            title = ''
            if 'Series' in parsed_json:
                title += parsed_json['Series']
            if 'GroupName' in parsed_json:
                title += '.' + parsed_json['GroupName']
            if 'Resolution' in parsed_json:
                title += '.' + parsed_json['Resolution']
            if 'Source' in parsed_json:
                title += '.' + parsed_json['Source']
            if 'Codec' in parsed_json:
                title += '.' + parsed_json['Codec']
            if title:
                title = title.replace(' ', '.')

        url = parsed_json.get('DownloadURL')
        if not url:
            sickrage.app.log.debug('Download URL is missing from response for release "{}"'.format(title))
        else:
            url = url.replace('\\/', '/')

        return title, url

    def _search_params(self, series_id, series_provider_id, season, episode, mode, season_numbering=None):
        searches = []

        show_object = find_show(series_id, series_provider_id)
        if not show_object:
            return searches

        episode_object = show_object.get_episode(season, episode)

        if not season_numbering and show_object.search_format in [SearchFormat.AIR_BY_DATE, SearchFormat.SPORTS]:
            date_fmt = '%Y' if mode == 'Season' else '%Y.%m.%d'
            search_name = episode_object.airdate.strftime(date_fmt)
        else:
            search_name = '{type} {number}'.format(
                type='Season' if mode == 'Season' else '',
                number=episode_object.get_season_episode_numbering()[0] if mode == 'Season' and season else
                episode_num(episode_object.get_season_episode_numbering()[0], episode_object.get_season_episode_numbering()[1]),
            ).strip()

        params = {
            'category': mode,
            'name': search_name,
        }

        # Search
        if show_object.series_provider_id == SeriesProviderID.THETVDB:
            params['tvdb'] = show_object.series_id
            params['series'] = show_object.name
            searches.append(params)

        for scene_exception in [x.split('|')[0] for x in show_object.scene_exceptions]:
            series_params = params.copy()
            series_params['series'] = sanitize_scene_name(scene_exception)
            searches.append(series_params)

        # extend air by date searches to include season numbering
        if show_object.search_format == SearchFormat.AIR_BY_DATE and not season_numbering:
            searches.extend(self._search_params(series_id, series_provider_id, season, episode, mode, season_numbering=True))

        return searches

    def _api_call(self, params=None, results_per_page=300, offset=0):
        response = {}

        json_rpc = {
            "jsonrpc": "2.0",
            "method": "getTorrents",
            "params": [self.custom_settings['api_key'], params or {}, results_per_page, offset],
            "id": uuid.uuid4().hex,
        }

        try:
            response = self.session.post(self.urls['api'], json=json_rpc, headers={'Content-Type': 'application/json-rpc'}).json()
            if response and 'error' in response:
                error = response["error"]
                message = error["message"]
                code = error["code"]
                if code == -32001:
                    sickrage.app.log.warning('Incorrect authentication credentials.')
                elif code == -32002:
                    sickrage.app.log.warning('You have exceeded the limit of 150 calls per hour.')
                elif code in (500, 502, 521, 524):
                    sickrage.app.log.warning('Provider is currently unavailable. Error: {} {}'.format(code, message))
                else:
                    sickrage.app.log.error('JSON-RPC protocol error while accessing provider. Error: {error!r}'.format(error=error))
            elif response and 'result' in response:
                response = response['result']
        except Exception as e:
            sickrage.app.log.warning("Error while accessing provider. Error: {}".format(e))

        return response



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
from marshmallow import fields

from sickrage.core.webserver.handlers.api.schemas import BaseSchema, BaseSuccessSchema


class SeriesSlugPath(BaseSchema):
    """Series slug schema"""

    series_slug = fields.String(
        required=True,
        default=False,
        description="Series slug for series you want to lookup",
    )


class EpisodeSlugPath(BaseSchema):
    """Episode slug schema"""

    episode_slug = fields.String(
        required=True,
        default=False,
        description="Episode slug for episode you want to lookup",
    )


class SeriesEpisodesRenameSuccessSchema(BaseSuccessSchema):
    pass


class SeriesEpisodesManualSearchPath(BaseSchema):
    """Complete episode manual search schema"""

    episode_slug = fields.String(
        required=True,
        default=False,
        description="Episode slug for episode you want to manually search for",
    )


class SeriesEpisodesManualSearchSchema(BaseSchema):
    """Complete episode manual search schema"""

    useExistingQuality = fields.Boolean(
        required=False,
        default=False,
        description="Use existing quality of previous manual episode search",
    )


class SeriesEpisodesManualSearchSuccessSchema(BaseSuccessSchema):
    pass



from __future__ import absolute_import, division, print_function

import requests
from trakt.interfaces.base import Interface, authenticated
from trakt.mapper.summary import SummaryMapper


class RecommendationsInterface(Interface):
    path = 'recommendations'

    @authenticated
    def shows(self, extended=None, **kwargs):
        response = self.http.get('shows', query={
            'extended': extended
        })

        items = self.get_data(response, **kwargs)

        if isinstance(items, requests.Response):
            return items

        return SummaryMapper.shows(self.client, items)


#!/usr/bin/env python
# coding=utf-8
#
# This file is part of aDBa.
#
# aDBa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# aDBa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with aDBa.  If not, see <http://www.gnu.org/licenses/>.

from threading import Lock

from .aniDBerrors import *
from .aniDBresponses import *


class Command:
    queue = {None: None}

    def __init__(self, command, **parameters):
        self.command = command
        self.parameters = parameters
        self.raw = self.flatten(command, parameters)

        self.mode = None
        self.callback = None
        self.waiter = Lock()
        self.waiter.acquire()

    def __repr__(self):
        return "Command(%r,%r) %r\n%s\n" % (self.tag, self.command, self.parameters, self.raw_data())

    def authorize(self, mode, tag, session, callback):
        self.mode = mode
        self.callback = callback
        self.tag = tag
        self.session = session

        self.parameters['tag'] = tag
        self.parameters['s'] = session

    def handle(self, resp):
        self.resp = resp
        if self.mode == 1:
            self.waiter.release()
        elif self.mode == 2:
            self.callback(resp)

    def wait_response(self):
        self.waiter.acquire()

    def flatten(self, command, parameters):
        tmp = []
        for key, value in parameters.items():
            if value is None:
                continue
            tmp.append("%s=%s" % (self.escape(key), self.escape(value)))
        return ' '.join([command, '&'.join(tmp)])

    @staticmethod
    def escape(data):
        return str(data).replace('&', '&amp;')

    def raw_data(self):
        self.raw = self.flatten(self.command, self.parameters)
        return self.raw

    def cached(self, interface, database):
        return None

    def cache(self, interface, database):
        pass


# first run
class AuthCommand(Command):
    def __init__(self, username, password, protover, client, clientver, nat=None, comp=None, enc=None, mtu=None):
        parameters = {'user': username, 'pass': password, 'protover': protover, 'client': client, 'clientver': clientver, 'nat': nat, 'comp': comp, 'enc': enc, 'mtu': mtu}
        Command.__init__(self, 'AUTH', **parameters)


class LogoutCommand(Command):
    def __init__(self):
        Command.__init__(self, 'LOGOUT')


# third run (at the same time as second)
class PushCommand(Command):
    def __init__(self, notify, msg, buddy=None):
        parameters = {'notify': notify, 'msg': msg, 'buddy': buddy}
        Command.__init__(self, 'PUSH', **parameters)


class PushAckCommand(Command):
    def __init__(self, nid):
        parameters = {'nid': nid}
        Command.__init__(self, 'PUSHACK', **parameters)


class NotifyAddCommand(Command):
    def __init__(self, aid=None, gid=None, type=None, priority=None):
        if not (aid or gid) or (aid and gid):
            raise AniDBIncorrectParameterError("You must provide aid OR gid for NOTIFICATIONADD command")
        parameters = {'aid': aid, "gid": gid, "type": type, "priority": priority}
        Command.__init__(self, 'NOTIFICATIONADD', **parameters)


class NotifyCommand(Command):
    def __init__(self, buddy=None):
        parameters = {'buddy': buddy}
        Command.__init__(self, 'NOTIFY', **parameters)


class NotifyListCommand(Command):
    def __init__(self):
        Command.__init__(self, 'NOTIFYLIST')


class NotifyGetCommand(Command):
    def __init__(self, type, id):
        parameters = {'type': type, 'id': id}
        Command.__init__(self, 'NOTIFYGET', **parameters)


class NotifyAckCommand(Command):
    def __init__(self, type, id):
        parameters = {'type': type, 'id': id}
        Command.__init__(self, 'NOTIFYACK', **parameters)


class BuddyAddCommand(Command):
    def __init__(self, uid=None, uname=None):
        if not (uid or uname) or (uid and uname):
            raise AniDBIncorrectParameterError("You must provide <u(id|name)> for BUDDYADD command")
        parameters = {'uid': uid, 'uname': uname.lower()}
        Command.__init__(self, 'BUDDYADD', **parameters)


class BuddyDelCommand(Command):
    def __init__(self, uid):
        parameters = {'uid': uid}
        Command.__init__(self, 'BUDDYDEL', **parameters)


class BuddyAcceptCommand(Command):
    def __init__(self, uid):
        parameters = {'uid': uid}
        Command.__init__(self, 'BUDDYACCEPT', **parameters)


class BuddyDenyCommand(Command):
    def __init__(self, uid):
        parameters = {'uid': uid}
        Command.__init__(self, 'BUDDYDENY', **parameters)


class BuddyListCommand(Command):
    def __init__(self, startat):
        parameters = {'startat': startat}
        Command.__init__(self, 'BUDDYLIST', **parameters)


class BuddyStateCommand(Command):
    def __init__(self, startat):
        parameters = {'startat': startat}
        Command.__init__(self, 'BUDDYSTATE', **parameters)


# first run
class AnimeCommand(Command):
    def __init__(self, aid=None, aname=None, amask=None):
        if not (aid or aname):
            raise AniDBIncorrectParameterError("You must provide <a(id|name)> for ANIME command")
        parameters = {'aid': aid, 'aname': aname, 'amask': amask}
        Command.__init__(self, 'ANIME', **parameters)


class EpisodeCommand(Command):
    def __init__(self, eid=None, aid=None, aname=None, epno=None):
        if not (eid or ((aname or aid) and epno)) or (aname and aid) or (eid and (aname or aid or epno)):
            raise AniDBIncorrectParameterError("You must provide <eid XOR a(id|name)+epno> for EPISODE command")
        parameters = {'eid': eid, 'aid': aid, 'aname': aname, 'epno': epno}
        Command.__init__(self, 'EPISODE', **parameters)


class FileCommand(Command):
    def __init__(self, fid=None, size=None, ed2k=None, aid=None, aname=None, gid=None, gname=None, epno=None, fmask=None, amask=None):
        if not (fid or (size and ed2k) or ((aid or aname) and (gid or gname) and epno)) or (fid and (size or ed2k or aid or aname or gid or gname or epno)) or ((size and ed2k) and (fid or aid or aname or gid or gname or epno)) or (((aid or aname) and (gid or gname) and epno) and (fid or size or ed2k)) or (aid and aname) or (gid and gname):
            raise AniDBIncorrectParameterError("You must provide <fid XOR size+ed2k XOR a(id|name)+g(id|name)+epno> for FILE command")
        parameters = {'fid': fid, 'size': size, 'ed2k': ed2k, 'aid': aid, 'aname': aname, 'gid': gid, 'gname': gname, 'epno': epno, 'fmask': fmask, 'amask': amask}
        Command.__init__(self, 'FILE', **parameters)


class GroupCommand(Command):
    def __init__(self, gid=None, gname=None):
        if not (gid or gname) or (gid and gname):
            raise AniDBIncorrectParameterError("You must provide <g(id|name)> for GROUP command")
        parameters = {'gid': gid, 'gname': gname}
        Command.__init__(self, 'GROUP', **parameters)


class GroupstatusCommand(Command):
    def __init__(self, aid=None, status=None):
        if not aid:
            raise AniDBIncorrectParameterError("You must provide aid for GROUPSTATUS command")
        parameters = {'aid': aid, 'status': status}
        Command.__init__(self, 'GROUPSTATUS', **parameters)


class ProducerCommand(Command):
    def __init__(self, pid=None, pname=None):
        if not (pid or pname) or (pid and pname):
            raise AniDBIncorrectParameterError("You must provide <p(id|name)> for PRODUCER command")
        parameters = {'pid': pid, 'pname': pname}
        Command.__init__(self, 'PRODUCER', **parameters)

    def cached(self, intr, db):
        pid = self.parameters['pid']
        pname = self.parameters['pname']

        codes = ('pid', 'name', 'shortname', 'othername', 'type', 'pic', 'url')
        names = ','.join([code for code in codes if code != ''])
        ruleholder = (pid and 'pid=%s' or '(name=%s OR shortname=%s OR othername=%s)')
        rulevalues = (pid and [pid] or [pname, pname, pname])

        rows = db.select('ptb', names, ruleholder + " AND status&8", *rulevalues)

        if len(rows) > 1:
            raise AniDBInternalError("It shouldn't be possible for database to return more than 1 line for PRODUCER cache")
        elif not len(rows):
            return None
        else:
            resp = ProducerResponse(self, None, '245', 'CACHED PRODUCER', [list(rows[0])])
            resp.parse()
            return resp

    def cache(self, intr, db):
        if self.resp.rescode != '245' or self.cached(intr, db):
            return

        codes = ('pid', 'name', 'shortname', 'othername', 'type', 'pic', 'url')
        if len(db.select('ptb', 'pid', 'pid=%s', self.resp.datalines[0]['pid'])):
            sets = 'status=status|15,' + ','.join([code + '=%s' for code in codes if code != ''])
            values = [self.resp.datalines[0][code] for code in codes if code != ''] + [self.resp.datalines[0]['pid']]

            db.update('ptb', sets, 'pid=%s', *values)
        else:
            names = 'status,' + ','.join([code for code in codes if code != ''])
            valueholders = '0,' + ','.join(['%s' for code in codes if code != ''])
            values = [self.resp.datalines[0][code] for code in codes if code != '']

            db.insert('ptb', names, valueholders, *values)


class MyListCommand(Command):
    def __init__(self, lid=None, fid=None, size=None, ed2k=None, aid=None, aname=None, gid=None, gname=None, epno=None):
        if not (lid or fid or (size and ed2k) or (aid or aname)) or (lid and (fid or size or ed2k or aid or aname or gid or gname or epno)) or (fid and (lid or size or ed2k or aid or aname or gid or gname or epno)) or ((size and ed2k) and (lid or fid or aid or aname or gid or gname or epno)) or ((aid or aname) and (lid or fid or size or ed2k)) or (aid and aname) or (gid and gname):
            raise AniDBIncorrectParameterError("You must provide <lid XOR fid XOR size+ed2k XOR a(id|name)+g(id|name)+epno> for MYLIST command")
        parameters = {'lid': lid, 'fid': fid, 'size': size, 'ed2k': ed2k, 'aid': aid, 'aname': aname, 'gid': gid, 'gname': gname, 'epno': epno}
        Command.__init__(self, 'MYLIST', **parameters)

    def cached(self, intr, db):
        lid = self.parameters['lid']
        fid = self.parameters['fid']
        size = self.parameters['size']
        ed2k = self.parameters['ed2k']
        aid = self.parameters['aid']
        aname = self.parameters['aname']
        gid = self.parameters['gid']
        gname = self.parameters['gname']
        epno = self.parameters['epno']

        names = ','.join([code for code in MylistResponse(None, None, None, None, []).codetail if code != ''])

        if lid:
            ruleholder = "lid=%s"
            rulevalues = [lid]
        elif fid or size or ed2k:
            resp = intr.file(fid=fid, size=size, ed2k=ed2k)
            if resp.rescode != '220':
                resp = NoSuchMylistFileResponse(self, None, '321', 'NO SUCH ENTRY (FILE NOT FOUND)', [])
                resp.parse()
                return resp
            fid = resp.datalines[0]['fid']

            ruleholder = "fid=%s"
            rulevalues = [fid]
        else:
            resp = intr.anime(aid=aid, aname=aname)
            if resp.rescode != '230':
                resp = NoSuchFileResponse(self, None, '321', 'NO SUCH ENTRY (ANIME NOT FOUND)', [])
                resp.parse()
                return resp
            aid = resp.datalines[0]['aid']

            resp = intr.group(gid=gid, gname=gname)
            if resp.rescode != '250':
                resp = NoSuchFileResponse(self, None, '321', 'NO SUCH ENTRY (GROUP NOT FOUND)', [])
                resp.parse()
                return resp
            gid = resp.datalines[0]['gid']

            resp = intr.episode(aid=aid, epno=epno)
            if resp.rescode != '240':
                resp = NoSuchFileResponse(self, None, '321', 'NO SUCH ENTRY (EPISODE NOT FOUND)', [])
                resp.parse()
                return resp
            eid = resp.datalines[0]['eid']

            ruleholder = "aid=%s AND eid=%s AND gid=%s"
            rulevalues = [aid, eid, gid]

        rows = db.select('ltb', names, ruleholder + " AND status&8", *rulevalues)

        if len(rows) > 1:
            # resp=MultipleFilesFoundResponse(self,None,'322','CACHED MULTIPLE FILES FOUND',/*get fids from rows, not gonna do this as you haven't got a real cache out of these..*/)
            return None
        elif not len(rows):
            return None
        else:
            resp = MylistResponse(self, None, '221', 'CACHED MYLIST', [list(rows[0])])
            resp.parse()
            return resp

    def cache(self, intr, db):
        if self.resp.rescode != '221' or self.cached(intr, db):
            return

        codes = MylistResponse(None, None, None, None, []).codetail
        if len(db.select('ltb', 'lid', 'lid=%s', self.resp.datalines[0]['lid'])):
            sets = 'status=status|15,' + ','.join([code + '=%s' for code in codes if code != ''])
            values = [self.resp.datalines[0][code] for code in codes if code != ''] + [self.resp.datalines[0]['lid']]

            db.update('ltb', sets, 'lid=%s', *values)
        else:
            names = 'status,' + ','.join([code for code in codes if code != ''])
            valueholders = '15,' + ','.join(['%s' for code in codes if code != ''])
            values = [self.resp.datalines[0][code] for code in codes if code != '']

            db.insert('ltb', names, valueholders, *values)


class MyListAddCommand(Command):
    def __init__(self, lid=None, fid=None, size=None, ed2k=None, aid=None, aname=None, gid=None, gname=None, epno=None, edit=None, state=None, viewed=None, source=None, storage=None, other=None):
        if not (lid or fid or (size and ed2k) or ((aid or aname) and (gid or gname))) or (lid and (fid or size or ed2k or aid or aname or gid or gname or epno)) or (fid and (lid or size or ed2k or aid or aname or gid or gname or epno)) or ((size and ed2k) and (lid or fid or aid or aname or gid or gname or epno)) or (((aid or aname) and (gid or gname)) and (lid or fid or size or ed2k)) or (aid and aname) or (gid and gname) or (lid and not edit):
            raise AniDBIncorrectParameterError("You must provide <lid XOR fid XOR size+ed2k XOR a(id|name)+g(id|name)+epno> for MYLISTADD command")
        parameters = {'lid': lid, 'fid': fid, 'size': size, 'ed2k': ed2k, 'aid': aid, 'aname': aname, 'gid': gid, 'gname': gname, 'epno': epno, 'edit': edit, 'state': state, 'viewed': viewed, 'source': source, 'storage': storage, 'other': other}
        Command.__init__(self, 'MYLISTADD', **parameters)


class MyListDelCommand(Command):
    def __init__(self, lid=None, fid=None, aid=None, aname=None, gid=None, gname=None, epno=None):
        if not (lid or fid or ((aid or aname) and (gid or gname) and epno)) or (lid and (fid or aid or aname or gid or gname or epno)) or (fid and (lid or aid or aname or gid or gname or epno)) or (((aid or aname) and (gid or gname) and epno) and (lid or fid)) or (aid and aname) or (gid and gname):
            raise AniDBIncorrectParameterError("You must provide <lid+edit=1 XOR fid XOR a(id|name)+g(id|name)+epno> for MYLISTDEL command")
        parameters = {'lid': lid, 'fid': fid, 'aid': aid, 'aname': aname, 'gid': gid, 'gname': gname, 'epno': epno}
        Command.__init__(self, 'MYLISTDEL', **parameters)


class MyListStatsCommand(Command):
    def __init__(self):
        Command.__init__(self, 'MYLISTSTATS')


class VoteCommand(Command):
    def __init__(self, type, id=None, name=None, value=None, epno=None):
        if not (id or name) or (id and name):
            raise AniDBIncorrectParameterError("You must provide <(id|name)> for VOTE command")
        parameters = {'type': type, 'id': id, 'name': name, 'value': value, 'epno': epno}
        Command.__init__(self, 'VOTE', **parameters)


class RandomAnimeCommand(Command):
    def __init__(self, type):
        parameters = {'type': type}
        Command.__init__(self, 'RANDOMANIME', **parameters)


class PingCommand(Command):
    def __init__(self):
        Command.__init__(self, 'PING')


# second run
class EncryptCommand(Command):
    def __init__(self, user, apipassword, type):
        self.apipassword = apipassword
        parameters = {'user': user.lower(), 'type': type}
        Command.__init__(self, 'ENCRYPT', **parameters)


class EncodingCommand(Command):
    def __init__(self, name):
        parameters = {'name': type}
        Command.__init__(self, 'ENCODING', **parameters)


class SendMsgCommand(Command):
    def __init__(self, to, title, body):
        if len(title) > 50 or len(body) > 900:
            raise AniDBIncorrectParameterError("Title must not be longer than 50 chars and body must not be longer than 900 chars for SENDMSG command")
        parameters = {'to': to.lower(), 'title': title, 'body': body}
        Command.__init__(self, 'SENDMSG', **parameters)


class UserCommand(Command):
    def __init__(self, user):
        parameters = {'user': user}
        Command.__init__(self, 'USER', **parameters)


class UptimeCommand(Command):
    def __init__(self):
        Command.__init__(self, 'UPTIME')


class VersionCommand(Command):
    def __init__(self):
        Command.__init__(self, 'VERSION')



# Author: Mr_Orange
# URL: https://github.com/mr-orange/Sick-Beard
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


from urllib.parse import urljoin

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import bs4_parser, try_int, convert_size
from sickrage.search_providers import TorrentProvider


class SpeedCDProvider(TorrentProvider):
    def __init__(self):
        super(SpeedCDProvider, self).__init__("Speedcd", 'https://speed.cd', True)

        self._urls.update({
            'login': '{base_url}/login.php'.format(**self._urls),
            'search': '{base_url}/browse.php'.format(**self._urls),
        })

        # custom settings
        self.custom_settings = {
            # 'username': '',
            # 'password': '',
            'freeleech': False,
            'minseed': 0,
            'minleech': 0
        }

        self.enable_cookies = True
        self.required_cookies = ('inSpeed_uid', 'inSpeed_speedian')

        self.proper_strings = ['PROPER', 'REPACK', 'REAL', 'RERIP']

        self.cache = TVCache(self, min_time=20)

    # def login(self):
    #     return self.cookie_login('log in')

    def login(self):
        return self.cookie_login('loginform')

        # if any(dict_from_cookiejar(self.session.cookies).values()):
        #     return True
        #
        # login_params = {
        #     'username': self.username,
        #     'password': self.password
        # }
        #
        # try:
        #     with bs4_parser(self.session.get(self.urls['login']).text) as html:
        #         login_url = urljoin(self.urls['base_url'], html.find('form', id='loginform').get('action'))
        #         response = self.session.post(login_url, data=login_params, timeout=30).text
        # except Exception as e:
        #     sickrage.app.log.warning("Unable to connect to provider")
        #     self.session.cookies.clear()
        #     return False
        #
        # if 'logout.php' not in response.lower():
        #     sickrage.app.log.warning("Invalid username or password, check your settings.")
        #     self.session.cookies.clear()
        #     return False
        #
        # return True

    def search(self, search_strings, age=0, series_id=None, series_provider_id=None, season=None, episode=None, **kwargs):
        results = []

        if not self.login():
            return results

        # http://speed.cd/browse.php?c49=1&c50=1&c52=1&c41=1&c55=1&c2=1&c30=1&freeleech=on&search=arrow&d=on
        # Search Params
        search_params = {
            'c2': 1,  # TV/Episodes
            'c30': 1,  # Anime
            'c41': 1,  # TV/Packs
            'c49': 1,  # TV/HD
            'c50': 1,  # TV/Sports
            'c52': 1,  # TV/B-Ray
            'c55': 1,  # TV/Kids
            'search': '',
            'freeleech': 'on' if self.custom_settings['freeleech'] else None
        }

        for mode in search_strings:
            sickrage.app.log.debug("Search Mode: %s" % mode)
            for search_string in search_strings[mode]:

                if mode != 'RSS':
                    sickrage.app.log.debug("Search string: %s " % search_string)

                search_params['search'] = search_string

                resp = self.session.get(self.urls['search'], params=search_params)
                if not resp or not resp.text:
                    sickrage.app.log.debug("No data returned from provider")
                    continue

                results += self.parse(resp.text, mode)

        return results

    def parse(self, data, mode, **kwargs):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        with bs4_parser(data) as html:
            torrent_table = html.find('div', class_='boxContent')
            torrent_table = torrent_table.find('table') if torrent_table else None
            torrent_rows = torrent_table('tr') if torrent_table else []

            # Continue only if at least one release is found
            if len(torrent_rows) < 2:
                sickrage.app.log.debug('Data returned from provider does not contain any torrents')
                return results

            # Skip column headers
            for row in torrent_rows[1:]:
                cells = row('td')

                try:
                    title = cells[1].find('a').get_text()
                    download_url = urljoin(self.urls['base_url'], cells[2].find(title='Download').parent['href'])
                    if not all([title, download_url]):
                        continue

                    seeders = try_int(cells[6].get_text(strip=True))
                    leechers = try_int(cells[7].get_text(strip=True))

                    torrent_size = cells[4].get_text()
                    torrent_size = torrent_size[:-2] + ' ' + torrent_size[-2:]
                    size = convert_size(torrent_size, -1)

                    results += [{
                        'title': title,
                        'link': download_url,
                        'size': size,
                        'seeders': seeders,
                        'leechers': leechers
                    }]

                    if mode != 'RSS':
                        sickrage.app.log.debug("Found result: {}".format(title))

                except (AttributeError, TypeError, KeyError, ValueError, IndexError):
                    sickrage.app.log.error('Failed parsing provider.')

        return results



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import threading

import sickrage
from sickrage.core.common import Quality, Qualities, EpisodeStatus
from sickrage.core.queues.search import DailySearchTask
from sickrage.core.tv.show.helpers import get_show_list


class DailySearcher(object):
    def __init__(self):
        self.name = "DAILYSEARCHER"
        self.lock = threading.Lock()
        self.running = False

    def task(self, force=False):
        """
        Runs the daily searcher, queuing selected episodes for search
        :param force: Force search
        """
        if self.running and not force:
            return

        try:
            self.running = True

            # set thread name
            threading.currentThread().setName(self.name)

            # find new released episodes and update their statuses
            for curShow in get_show_list():
                if curShow.paused:
                    sickrage.app.log.debug("Skipping search for {} because the show is paused".format(curShow.name))
                    continue

                for tv_episode in curShow.new_episodes:
                    tv_episode.status = tv_episode.show.default_ep_status if tv_episode.season > 0 else EpisodeStatus.SKIPPED
                    tv_episode.save()
                    sickrage.app.log.info('Setting status ({status}) for show airing today: {name} {special}'.format(
                        name=tv_episode.pretty_name(),
                        status=tv_episode.status.display_name,
                        special='(specials are not supported)' if not tv_episode.season > 0 else '',
                    ))

                wanted = self._get_wanted(curShow, datetime.date.today())
                if not wanted:
                    sickrage.app.log.debug("Nothing needs to be downloaded for {}, skipping".format(curShow.name))
                    continue

                for season, episode in wanted:
                    if (curShow.series_id, season, episode) in sickrage.app.search_queue.SNATCH_HISTORY:
                        sickrage.app.search_queue.SNATCH_HISTORY.remove((curShow.series_id, season, episode))

                    sickrage.app.search_queue.put(DailySearchTask(curShow.series_id, curShow.series_provider_id, season, episode))
        finally:
            self.running = False

    @staticmethod
    def _get_wanted(show, from_date):
        """
        Get a list of episodes that we want to download
        :param show: Show these episodes are from
        :param fromDate: Search from a certain date
        :return: list of wanted episodes
        """

        wanted = []

        any_qualities, best_qualities = Quality.split_quality(show.quality)
        all_qualities = list(set(any_qualities + best_qualities))

        sickrage.app.log.debug("Seeing if we need anything for today from {}".format(show.name))

        # check through the list of statuses to see if we want any
        for episode_object in show.episodes:
            if not episode_object.season > 0 or not episode_object.airdate >= from_date:
                continue

            cur_status, cur_quality = Quality.split_composite_status(episode_object.status)

            # if we need a better one then say yes
            if cur_status not in (EpisodeStatus.WANTED, EpisodeStatus.DOWNLOADED, EpisodeStatus.SNATCHED, EpisodeStatus.SNATCHED_PROPER):
                continue

            if cur_status != EpisodeStatus.WANTED:
                if best_qualities:
                    if cur_quality in best_qualities:
                        continue
                    elif cur_quality != Qualities.UNKNOWN and cur_quality > max(best_qualities):
                        continue
                elif any_qualities:
                    if cur_quality in any_qualities:
                        continue
                    elif cur_quality != Qualities.UNKNOWN and cur_quality > max(any_qualities):
                        continue

            # skip upgrading quality of downloaded episodes if enabled
            if cur_status == EpisodeStatus.DOWNLOADED and show.skip_downloaded:
                continue

            wanted += [(episode_object.season, episode_object.episode)]

        return wanted



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################
import enum


class SeriesProviderID(enum.Enum):
    THETVDB = 'thetvdb'

    @property
    def _strings(self):
        return {
            self.THETVDB.name: 'TheTVDB'
        }

    @property
    def _slug_strings(self):
        return {
            self.THETVDB.name: 'thetvdb'
        }

    @property
    def display_name(self):
        return self._strings[self.name]

    @property
    def slug(self):
        return self._slug_strings[self.name]

    @classmethod
    def by_slug(cls, value):
        for item in cls:
            if item.slug == value:
                return item


class DefaultHomePage(enum.Enum):
    HOME = 'home'
    SCHEDULE = 'schedule'
    HISTORY = 'history'

    @property
    def _strings(self):
        return {
            self.HOME.name: 'Home',
            self.SCHEDULE.name: 'Schedule',
            self.HISTORY.name: 'History',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class MultiEpNaming(enum.Enum):
    REPEAT = 1
    EXTEND = 2
    DUPLICATE = 4
    LIMITED_EXTEND = 8
    SEPARATED_REPEAT = 16
    LIMITED_EXTEND_E_PREFIXED = 32

    @property
    def _strings(self):
        return {
            self.REPEAT.name: 'Repeat',
            self.SEPARATED_REPEAT.name: 'Repeat (Separated)',
            self.DUPLICATE.name: 'Duplicate',
            self.EXTEND.name: 'Extend',
            self.LIMITED_EXTEND.name: 'Extend (Limited)',
            self.LIMITED_EXTEND_E_PREFIXED.name: 'Extend (Limited, E-prefixed)'
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class CpuPreset(enum.Enum):
    LOW = 0.01
    NORMAL = 0.02
    HIGH = 0.05

    @property
    def _strings(self):
        return {
            self.LOW.name: 'Low',
            self.NORMAL.name: 'Normal',
            self.HIGH.name: 'High',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class CheckPropersInterval(enum.Enum):
    DAILY = 24 * 60
    FOUR_HOURS = 4 * 60
    NINETY_MINUTES = 90
    FORTY_FIVE_MINUTES = 45
    FIFTEEN_MINUTES = 15

    @property
    def _strings(self):
        return {
            self.DAILY.name: '24 hours',
            self.FOUR_HOURS.name: '4 hours',
            self.NINETY_MINUTES.name: '90 mins',
            self.FORTY_FIVE_MINUTES.name: '45 mins',
            self.FIFTEEN_MINUTES.name: '15 mins',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class FileTimestampTimezone(enum.Enum):
    NETWORK = 0
    LOCAL = 1

    @property
    def _strings(self):
        return {
            self.NETWORK.name: 'Network',
            self.LOCAL.name: 'Local',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class ProcessMethod(enum.Enum):
    COPY = 'copy'
    MOVE = 'move'
    HARDLINK = 'hardlink'
    SYMLINK = 'symlink'
    SYMLINK_REVERSED = 'symlink_reversed'

    @property
    def _strings(self):
        return {
            self.COPY.name: 'Copy',
            self.MOVE.name: 'Move',
            self.HARDLINK.name: 'Hard Link',
            self.SYMLINK.name: 'Symbolic Link',
            self.SYMLINK_REVERSED.name: 'Symbolic Link Reversed',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class NzbMethod(enum.Enum):
    BLACKHOLE = 'blackhole'
    SABNZBD = 'sabnzbd'
    NZBGET = 'nzbget'
    DOWNLOAD_STATION = 'download_station'

    @property
    def _strings(self):
        return {
            self.BLACKHOLE.name: 'Blackhole',
            self.SABNZBD.name: 'SABnzbd',
            self.NZBGET.name: 'NZBget',
            self.DOWNLOAD_STATION.name: 'Synology DS',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class TorrentMethod(enum.Enum):
    BLACKHOLE = 'blackhole'
    UTORRENT = 'utorrent'
    TRANSMISSION = 'transmission'
    DELUGE = 'deluge'
    DELUGED = 'deluged'
    DOWNLOAD_STATION = 'download_station'
    RTORRENT = 'rtorrent'
    QBITTORRENT = 'qbittorrent'
    MLNET = 'mlnet'
    PUTIO = 'putio'

    @property
    def _strings(self):
        return {
            self.BLACKHOLE.name: 'Blackhole',
            self.UTORRENT.name: 'uTorrent',
            self.TRANSMISSION.name: 'Transmission',
            self.DELUGE.name: 'Deluge (via WebUI)',
            self.DELUGED.name: 'Deluge (via Daemon)',
            self.DOWNLOAD_STATION.name: 'Synology DS',
            self.RTORRENT.name: 'rTorrent',
            self.QBITTORRENT.name: 'qBitTorrent',
            self.MLNET.name: 'MLDonkey',
            self.PUTIO.name: 'Putio',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class SearchFormat(enum.Enum):
    STANDARD = 1
    AIR_BY_DATE = 2
    ANIME = 3
    SPORTS = 4
    COLLECTION = 6

    @property
    def _strings(self):
        return {
            self.STANDARD.name: 'Standard (Show.S01E01)',
            self.AIR_BY_DATE.name: 'Air By Date (Show.2010.03.02)',
            self.ANIME.name: 'Anime (Show.265)',
            self.SPORTS.name: 'Sports (Show.2010.03.02)',
            self.COLLECTION.name: 'Collection (Show.Series.1.1of10) or (Show.Series.1.Part.1)'
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class UserPermission(enum.Enum):
    SUPERUSER = 0
    GUEST = 1

    @property
    def _strings(self):
        return {
            self.SUPERUSER.name: 'Superuser',
            self.GUEST.name: 'Guest',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class PosterSortDirection(enum.Enum):
    ASCENDING = 0
    DESCENDING = 1

    @property
    def _strings(self):
        return {
            self.ASCENDING.name: 'Ascending',
            self.DESCENDING.name: 'Descending',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class HomeLayout(enum.Enum):
    POSTER = 'poster'
    SMALL = 'small'
    BANNER = 'banner'
    DETAILED = 'detailed'
    SIMPLE = 'simple'

    @property
    def _strings(self):
        return {
            self.POSTER.name: 'Poster',
            self.SMALL.name: 'Small Poster',
            self.BANNER.name: 'Banner',
            self.DETAILED.name: 'Detailed',
            self.SIMPLE.name: 'Simple',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class PosterSortBy(enum.Enum):
    NAME = 0
    DATE = 1
    NETWORK = 2
    PROGRESS = 3

    @property
    def _strings(self):
        return {
            self.NAME.name: 'Sort By Name',
            self.DATE.name: 'Sort By Date',
            self.NETWORK.name: 'Sort By Network',
            self.PROGRESS.name: 'Sort By Progress',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class HistoryLayout(enum.Enum):
    DETAILED = 'detailed'
    COMPACT = 'compact'

    @property
    def _strings(self):
        return {
            self.DETAILED.name: 'Detailed',
            self.COMPACT.name: 'Compact',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class TimezoneDisplay(enum.Enum):
    LOCAL = 0
    NETWORK = 1

    @property
    def _strings(self):
        return {
            self.LOCAL.name: 'Local',
            self.NETWORK.name: 'Network',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class UITheme(enum.Enum):
    DARK = 'dark'
    LIGHT = 'light'

    @property
    def _strings(self):
        return {
            self.DARK.name: 'Dark',
            self.LIGHT.name: 'Light',
        }

    @property
    def display_name(self):
        return self._strings[self.name]


class TraktAddMethod(enum.Enum):
    SKIP_ALL = 0
    DOWNLOAD_PILOT_ONLY = 1
    WHOLE_SHOW = 2

    @property
    def _strings(self):
        return {
            self.SKIP_ALL.name: 'Skip All',
            self.DOWNLOAD_PILOT_ONLY.name: 'Download Pilot Only',
            self.WHOLE_SHOW.name: 'Get Whole Show',
        }

    @property
    def display_name(self):
        return self._strings[self.name]



from __future__ import absolute_import, division, print_function

from trakt.core.helpers import popitems
from trakt.interfaces.base import Interface, authenticated


class UsersSettingsInterface(Interface):
    path = 'users/settings'

    @authenticated
    def get(self, **kwargs):
        response = self.http.get(
            **popitems(kwargs, [
                'authenticated',
                'validate_token'
            ])
        )

        return self.get_data(response, **kwargs)



"""Initial migration

Revision ID: 7
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from sickrage.core.enums import SeriesProviderID

revision = '9'
down_revision = '8'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    providers = sa.Table('providers', meta, autoload=True)

    op.add_column('providers', sa.Column('series_provider_id', sa.Enum, default=SeriesProviderID.THETVDB))

    with op.get_context().begin_transaction():
        for row in conn.execute(providers.select()):
            conn.execute(f'UPDATE providers SET series_provider_id = "{SeriesProviderID.THETVDB.name}" WHERE providers.id = {row.id}')


def downgrade():
    pass



# Author: Dieter Blomme <dieterblomme@gmail.com>
# URL: https://sickrage.ca
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import sickrage
from sickrage.core.traktapi import TraktAPI
from sickrage.notification_providers import NotificationProvider


class TraktNotification(NotificationProvider):
    """
    A "notifier" for trakt.tv which keeps track of what has and hasn't been added to your library.
    """

    def __init__(self):
        super(TraktNotification, self).__init__()
        self.name = 'trakt'

    def notify_snatch(self, ep_name):
        pass

    def notify_download(self, ep_name):
        pass

    def notify_subtitle_download(self, ep_name, lang):
        pass

    def notify_version_update(self, new_version):
        pass

    def update_library(self, ep_obj):
        """
        Sends a request to trakt indicating that the given episode is part of our library.

        ep_obj: The TVEpisode object to add to trakt
        """

        if sickrage.app.config.trakt.enable:
            try:
                # URL parameters
                data = {
                    'shows': [
                        {
                            'title': ep_obj.show.name,
                            'year': ep_obj.show.startyear,
                            'ids': {sickrage.app.series_provider[ep_obj.show.series_provider_id].trakt_id: ep_obj.show.series_id},
                        }
                    ]
                }

                if sickrage.app.config.trakt.sync_watchlist:
                    if sickrage.app.config.trakt.remove_serieslist:
                        TraktAPI()["sync/watchlist"].remove(data)

                # Add Season and Episode + Related Episodes
                data['shows'][0]['seasons'] = [{'number': ep_obj.season, 'episodes': []}]

                for relEp_Obj in [ep_obj] + ep_obj.related_episodes:
                    data['shows'][0]['seasons'][0]['episodes'].append({'number': relEp_Obj.episode})

                if sickrage.app.config.trakt.sync_watchlist:
                    if sickrage.app.config.trakt.remove_watchlist:
                        TraktAPI()["sync/watchlist"].remove(data)

                # update library
                TraktAPI()["sync/collection"].add(data)

            except Exception as e:
                sickrage.app.log.warning("Could not connect to Trakt service: %s" % e)

    def update_watchlist(self, show_object=None, s=None, e=None, data_show=None, data_episode=None, update="add"):

        """
        Sends a request to trakt indicating that the given episode is part of our library.

        show_obj: The TVShow object to add to trakt
        s: season number
        e: episode number
        data_show: structured object of shows traktv type
        data_episode: structured object of episodes traktv type
        update: type o action add or remove
        """

        sickrage.app.log.debug(f"Add episodes, series_id: {show_object.series_id}, Title {show_object.name} to Trak.tv Watchlist")

        if sickrage.app.config.trakt.enable:
            data = {}
            try:
                # URL parameters
                if show_object is not None:
                    data = {
                        'shows': [
                            {
                                'title': show_object.name,
                                'year': show_object.startyear,
                                'ids': {sickrage.app.series_provider[show_object.series_provider_id].trakt_id: show_object.series_id},
                            }
                        ]
                    }
                elif data_show is not None:
                    data.update(data_show)
                else:
                    sickrage.app.log.warning("there's a coding problem contact developer. It's needed to be provided "
                                             "at lest one of the two: data_show or show_obj")
                    return False

                if data_episode is not None:
                    data['shows'][0].update(data_episode)

                elif s is not None:
                    # traktv URL parameters
                    season = {
                        'season': [
                            {
                                'number': s,
                            }
                        ]
                    }

                    if e is not None:
                        # traktv URL parameters
                        episode = {
                            'episodes': [
                                {
                                    'number': e
                                }
                            ]
                        }

                        season['season'][0].update(episode)

                    data['shows'][0].update(season)

                trakt_url = "sync/watchlist"
                if update == "remove":
                    TraktAPI()[trakt_url].remove(data)
                else:
                    TraktAPI()[trakt_url].add(data)

            except Exception as e:
                sickrage.app.log.warning("Could not connect to Trakt service: %s" % e)
                return False

        return True

    def trakt_show_data_generate(self, data):
        show_list = []

        for series_provider_id, series_id, title, year in data:
            show = {'title': title,
                    'year': year,
                    'ids': {sickrage.app.series_provider[series_provider_id].trakt_id: series_id}}

            show_list.append(show)

        post_data = {'shows': show_list}

        return post_data

    def trakt_episode_data_generate(self, data):
        # Find how many unique season we have
        unique_seasons = []

        for season, episode in data:
            if season not in unique_seasons:
                unique_seasons.append(season)

        # build the query
        seasonsList = []
        for searchedSeason in unique_seasons:
            episodesList = []
            for season, episode in data:
                if season == searchedSeason:
                    episodesList.append({'number': episode})
            seasonsList.append({'number': searchedSeason, 'episodes': episodesList})

        post_data = {'seasons': seasonsList}

        return post_data

    def test_notify(self, username, blacklist_name=None):
        """
        Sends a test notification to trakt with the given authentication info and returns a boolean
        representing success.

        api: The api string to use
        username: The username to use
        blacklist_name: slug of trakt list used to hide not interested show

        Returns: True if the request succeeded, False otherwise
        """
        try:
            if blacklist_name and blacklist_name is not None:
                if not TraktAPI()["users/me/lists/{list}".format(list=blacklist_name)].get():
                    return "Trakt blacklist doesn't exists"
            return "Test notice sent successfully to Trakt"
        except Exception as e:
            sickrage.app.log.warning("Could not connect to Trakt service: %s" % e)
            return "Test notice failed to Trakt: %s" % e



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################



from tornado.escape import json_encode
from tornado.web import authenticated

import sickrage
from sickrage.core.config.helpers import change_subtitle_searcher_freq
from sickrage.core.helpers import checkbox_to_value
from sickrage.core.webserver import ConfigWebHandler
from sickrage.core.webserver.handlers.base import BaseHandler
from sickrage.subtitles import Subtitles


class ConfigSubtitlesHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return self.render('config/subtitles.mako',
                           submenu=ConfigWebHandler.menu,
                           title=_('Config - Subtitles Settings'),
                           header=_('Subtitles Settings'),
                           topmenu='config',
                           controller='config',
                           action='subtitles')


class ConfigSubtitleGetCodeHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        q = self.get_argument('q')

        codes = [{"id": code, "name": Subtitles().name_from_code(code)} for code in Subtitles().subtitle_code_filter()]
        codes = list(filter(lambda code: q.lower() in code['name'].lower(), codes))

        return json_encode(codes)


class ConfigSubtitlesWantedLanguagesHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        codes = [{"id": code, "name": Subtitles().name_from_code(code)} for code in Subtitles().subtitle_code_filter()]
        codes = list(filter(lambda code: code['id'] in Subtitles().wanted_languages(), codes))

        return json_encode(codes)


class SaveSubtitlesHandler(BaseHandler):
    @authenticated
    def post(self, *args, **kwargs):
        use_subtitles = self.get_argument('use_subtitles', None)
        subtitles_dir = self.get_argument('subtitles_dir', None)
        service_order = self.get_argument('service_order', None)
        subtitles_history = self.get_argument('subtitles_history', None)
        subtitles_finder_frequency = self.get_argument('subtitles_finder_frequency', None)
        subtitles_multi = self.get_argument('subtitles_multi', None)
        enable_embedded_subtitles = self.get_argument('enable_embedded_subtitles', None)
        subtitles_extra_scripts = self.get_argument('subtitles_extra_scripts', '')
        subtitles_hearing_impaired = self.get_argument('subtitles_hearing_impaired', None)
        itasa_user = self.get_argument('itasa_user', None)
        itasa_pass = self.get_argument('itasa_pass', None)
        addic7ed_user = self.get_argument('addic7ed_user', None)
        addic7ed_pass = self.get_argument('addic7ed_pass', None)
        legendastv_user = self.get_argument('legendastv_user', None)
        legendastv_pass = self.get_argument('legendastv_pass', None)
        opensubtitles_user = self.get_argument('opensubtitles_user', None)
        opensubtitles_pass = self.get_argument('opensubtitles_pass', None)
        subtitles_languages = self.get_arguments('subtitles_languages[]')

        results = []

        change_subtitle_searcher_freq(subtitles_finder_frequency)

        sickrage.app.config.subtitles.enable = checkbox_to_value(use_subtitles)
        sickrage.app.config.subtitles.dir = subtitles_dir
        sickrage.app.config.subtitles.history = checkbox_to_value(subtitles_history)
        sickrage.app.config.subtitles.enable_embedded = checkbox_to_value(enable_embedded_subtitles)
        sickrage.app.config.subtitles.hearing_impaired = checkbox_to_value(subtitles_hearing_impaired)
        sickrage.app.config.subtitles.multi = checkbox_to_value(subtitles_multi)
        sickrage.app.config.subtitles.extra_scripts = subtitles_extra_scripts

        # Subtitle languages
        sickrage.app.config.subtitles.languages = ','.join(subtitles_languages) or 'eng'

        # Subtitles services
        services_str_list = service_order.split()
        subtitles_services_list = []
        subtitles_services_enabled = []
        for curServiceStr in services_str_list:
            cur_service, cur_enabled = curServiceStr.split(':')
            subtitles_services_list.append(cur_service)
            subtitles_services_enabled.append(cur_enabled)

        sickrage.app.config.subtitles.services_list = ','.join(subtitles_services_list)
        sickrage.app.config.subtitles.services_enabled = '|'.join(subtitles_services_enabled)

        sickrage.app.config.subtitles.addic7ed_user = addic7ed_user or ''
        sickrage.app.config.subtitles.addic7ed_pass = addic7ed_pass or ''
        sickrage.app.config.subtitles.legendastv_user = legendastv_user or ''
        sickrage.app.config.subtitles.legendastv_pass = legendastv_pass or ''
        sickrage.app.config.subtitles.itasa_user = itasa_user or ''
        sickrage.app.config.subtitles.itasa_pass = itasa_pass or ''
        sickrage.app.config.subtitles.opensubtitles_user = opensubtitles_user or ''
        sickrage.app.config.subtitles.opensubtitles_pass = opensubtitles_pass or ''

        sickrage.app.config.save()

        if len(results) > 0:
            [sickrage.app.log.error(x) for x in results]
            sickrage.app.alerts.error(_('Error(s) Saving Configuration'), '<br>\n'.join(results))
        else:
            sickrage.app.alerts.message(_('[SUBTITLES] Configuration Saved to Database'))

        return self.redirect("/config/subtitles/")



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import os

from tornado.web import authenticated

import sickrage
from sickrage.core.common import Quality, Qualities, EpisodeStatus
from sickrage.core.config.helpers import change_gui_lang, change_https_key, change_https_cert, change_show_update_hour, \
    change_version_notify
from sickrage.core.enums import UITheme, DefaultHomePage, TimezoneDisplay, SearchFormat, SeriesProviderID, CpuPreset
from sickrage.core.helpers import generate_api_key, checkbox_to_value, try_int
from sickrage.core.webserver import ConfigWebHandler
from sickrage.core.webserver.handlers.base import BaseHandler


class ConfigGeneralHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return self.render('config/general.mako',
                           title=_('Config - General'),
                           header=_('General Configuration'),
                           topmenu='config',
                           submenu=ConfigWebHandler.menu,
                           controller='config',
                           action='general', )


class GenerateApiKeyHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return generate_api_key()


class SaveRootDirsHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        sickrage.app.config.general.root_dirs = self.get_argument('rootDirString', '')
        sickrage.app.config.save()


class SaveAddShowDefaultsHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        default_status = self.get_argument('defaultStatus', '5')
        quality_preset = self.get_argument('qualityPreset', '')
        any_qualities = self.get_argument('anyQualities', '')
        best_qualities = self.get_argument('bestQualities', '')
        default_flatten_folders = self.get_argument('defaultFlattenFolders', None)
        subtitles = self.get_argument('subtitles', None)
        anime = self.get_argument('anime', None)
        search_format = self.get_argument('search_format', None)
        default_status_after = self.get_argument('defaultStatusAfter', None) or EpisodeStatus.WANTED
        scene = self.get_argument('scene', None)
        skip_downloaded = self.get_argument('skip_downloaded', None)
        add_show_year = self.get_argument('add_show_year', None)

        any_qualities = any_qualities.split(',') if len(any_qualities) else []
        best_qualities = best_qualities.split(',') if len(best_qualities) else []

        try:
            new_quality = Qualities[quality_preset]
        except KeyError:
            new_quality = Quality.combine_qualities([Qualities[x] for x in any_qualities], [Qualities[x] for x in best_qualities])

        sickrage.app.config.general.status_default = EpisodeStatus[default_status]
        sickrage.app.config.general.status_default_after = EpisodeStatus[default_status_after]
        sickrage.app.config.general.quality_default = new_quality

        sickrage.app.config.general.flatten_folders_default = not checkbox_to_value(default_flatten_folders)
        sickrage.app.config.subtitles.default = checkbox_to_value(subtitles)

        sickrage.app.config.general.anime_default = checkbox_to_value(anime)
        sickrage.app.config.general.search_format_default = SearchFormat[search_format]
        sickrage.app.config.general.scene_default = checkbox_to_value(scene)
        sickrage.app.config.general.skip_downloaded_default = checkbox_to_value(skip_downloaded)
        sickrage.app.config.general.add_show_year_default = checkbox_to_value(add_show_year)

        sickrage.app.config.save()


class SaveGeneralHandler(BaseHandler):
    @authenticated
    def post(self, *args, **kwargs):
        log_nr = self.get_argument('log_nr', '5')
        log_size = self.get_argument('log_size', '1048576')
        web_port = self.get_argument('web_port', None)
        web_ipv6 = self.get_argument('web_ipv6', None)
        web_host = self.get_argument('web_host', None)
        trash_remove_show = self.get_argument('trash_remove_show', None)
        trash_rotate_logs = self.get_argument('trash_rotate_logs', None)
        update_frequency = self.get_argument('update_frequency', None)
        skip_removed_files = self.get_argument('skip_removed_files', None)
        series_provider_default_language = self.get_argument('series_provider_default_language', 'en')
        ep_default_deleted_status = self.get_argument('ep_default_deleted_status', None)
        launch_browser = self.get_argument('launch_browser', None)
        show_update_hour = self.get_argument('show_update_hour', '3')
        api_key = self.get_argument('api_key', None)
        series_provider_default = self.get_argument('series_provider_default', None)
        timezone_display = self.get_argument('timezone_display', None)
        cpu_preset = self.get_argument('cpu_preset', 'NORMAL')
        version_notify = self.get_argument('version_notify', None)
        enable_https = self.get_argument('enable_https', None)
        https_cert = self.get_argument('https_cert', None)
        https_key = self.get_argument('https_key', None)
        handle_reverse_proxy = self.get_argument('handle_reverse_proxy', None)
        sort_article = self.get_argument('sort_article', None)
        auto_update = self.get_argument('auto_update', None)
        notify_on_update = self.get_argument('notify_on_update', None)
        backup_on_update = self.get_argument('backup_on_update', None)
        proxy_setting = self.get_argument('proxy_setting', None)
        proxy_series_providers = self.get_argument('proxy_series_providers', None)
        anon_redirect = self.get_argument('anon_redirect', None)
        git_path = self.get_argument('git_path', None)
        pip3_path = self.get_argument('pip3_path', None)
        calendar_unprotected = self.get_argument('calendar_unprotected', None)
        calendar_icons = self.get_argument('calendar_icons', None)
        debug = self.get_argument('debug', None)
        ssl_verify = self.get_argument('ssl_verify', None)
        no_restart = self.get_argument('no_restart', None)
        coming_eps_missed_range = self.get_argument('coming_eps_missed_range', None)
        filter_row = self.get_argument('filter_row', None)
        fuzzy_dating = self.get_argument('fuzzy_dating', None)
        trim_zero = self.get_argument('trim_zero', None)
        date_preset = self.get_argument('date_preset', None)
        time_preset = self.get_argument('time_preset', None)
        series_provider_timeout = self.get_argument('series_provider_timeout', None)
        download_url = self.get_argument('download_url', None)
        theme_name = self.get_argument('theme_name', None)
        default_page = self.get_argument('default_page', None)
        gui_language = self.get_argument('gui_language', None)
        display_all_seasons = self.get_argument('display_all_seasons', None)
        show_update_stale = self.get_argument('show_update_stale', None)
        notify_on_login = self.get_argument('notify_on_login', None)
        allowed_video_file_exts = self.get_argument('allowed_video_file_exts', '')
        enable_upnp = self.get_argument('enable_upnp', None)
        strip_special_file_bits = self.get_argument('strip_special_file_bits', None)
        max_queue_workers = self.get_argument('max_queue_workers', None)
        web_root = self.get_argument('web_root', '')
        ip_whitelist_localhost_enabled = self.get_argument('ip_whitelist_localhost_enabled', None)
        ip_whitelist_enabled = self.get_argument('ip_whitelist_enabled', None)
        ip_whitelist = self.get_argument('ip_whitelist', '')
        web_auth_method = self.get_argument('web_auth_method', '')
        web_username = self.get_argument('web_username', '')
        web_password = self.get_argument('web_password', '')
        enable_sickrage_api = self.get_argument('enable_sickrage_api', None)

        results = []

        change_gui_lang(gui_language)
        change_show_update_hour(show_update_hour)
        change_version_notify(checkbox_to_value(version_notify))

        # Debug
        sickrage.app.config.general.debug = sickrage.app.debug = checkbox_to_value(debug)
        sickrage.app.log.set_level()

        # Misc
        sickrage.app.config.general.enable_upnp = checkbox_to_value(enable_upnp)
        sickrage.app.config.general.download_url = download_url
        sickrage.app.config.general.series_provider_default_language = series_provider_default_language
        sickrage.app.config.general.ep_default_deleted_status = EpisodeStatus[ep_default_deleted_status]
        sickrage.app.config.general.skip_removed_files = checkbox_to_value(skip_removed_files)
        sickrage.app.config.general.launch_browser = checkbox_to_value(launch_browser)
        sickrage.app.config.general.auto_update = checkbox_to_value(auto_update)
        sickrage.app.config.general.notify_on_update = checkbox_to_value(notify_on_update)
        sickrage.app.config.general.backup_on_update = checkbox_to_value(backup_on_update)
        sickrage.app.config.general.notify_on_login = checkbox_to_value(notify_on_login)
        sickrage.app.config.general.show_update_stale = checkbox_to_value(show_update_stale)
        sickrage.app.config.general.log_nr = int(log_nr)
        sickrage.app.config.general.log_size = int(log_size)

        sickrage.app.config.general.trash_remove_show = checkbox_to_value(trash_remove_show)
        sickrage.app.config.general.trash_rotate_logs = checkbox_to_value(trash_rotate_logs)
        sickrage.app.config.general.launch_browser = checkbox_to_value(launch_browser)
        sickrage.app.config.general.sort_article = checkbox_to_value(sort_article)
        sickrage.app.config.general.cpu_preset = CpuPreset[cpu_preset]
        sickrage.app.config.general.anon_redirect = anon_redirect
        sickrage.app.config.general.proxy_setting = proxy_setting
        sickrage.app.config.general.proxy_series_providers = checkbox_to_value(proxy_series_providers)
        sickrage.app.config.general.git_reset = 1
        sickrage.app.config.general.git_path = git_path
        sickrage.app.config.general.pip3_path = pip3_path
        sickrage.app.config.general.calendar_unprotected = checkbox_to_value(calendar_unprotected)
        sickrage.app.config.general.calendar_icons = checkbox_to_value(calendar_icons)
        sickrage.app.config.general.no_restart = checkbox_to_value(no_restart)

        sickrage.app.config.general.ssl_verify = checkbox_to_value(ssl_verify)
        sickrage.app.config.gui.coming_eps_missed_range = try_int(coming_eps_missed_range, 7)
        sickrage.app.config.general.display_all_seasons = checkbox_to_value(display_all_seasons)

        sickrage.app.config.general.web_port = int(web_port)
        sickrage.app.config.general.web_ipv6 = checkbox_to_value(web_ipv6)

        sickrage.app.config.gui.filter_row = checkbox_to_value(filter_row)
        sickrage.app.config.gui.fuzzy_dating = checkbox_to_value(fuzzy_dating)
        sickrage.app.config.gui.trim_zero = checkbox_to_value(trim_zero)

        sickrage.app.config.general.allowed_video_file_exts = ','.join([x.lower() for x in allowed_video_file_exts.split(',')])

        sickrage.app.config.general.strip_special_file_bits = checkbox_to_value(strip_special_file_bits)

        sickrage.app.config.general.web_root = web_root

        sickrage.app.config.general.ip_whitelist_enabled = checkbox_to_value(ip_whitelist_enabled)
        sickrage.app.config.general.ip_whitelist_localhost_enabled = checkbox_to_value(ip_whitelist_localhost_enabled)
        sickrage.app.config.general.ip_whitelist = ip_whitelist

        if web_auth_method == 'sso_auth':
            auth_method_changed = not sickrage.app.config.general.sso_auth_enabled
            sickrage.app.config.general.sso_auth_enabled = True
            sickrage.app.config.general.local_auth_enabled = False
        else:
            auth_method_changed = not sickrage.app.config.general.local_auth_enabled
            sickrage.app.config.general.sso_auth_enabled = False
            sickrage.app.config.general.local_auth_enabled = True
            sickrage.app.config.user.username = web_username
            sickrage.app.config.user.password = web_password

        sickrage.app.config.general.enable_sickrage_api = checkbox_to_value(enable_sickrage_api)

        # change_web_external_port(web_external_port)

        if date_preset:
            sickrage.app.config.gui.date_preset = date_preset

        if series_provider_default:
            sickrage.app.config.general.series_provider_default = SeriesProviderID[series_provider_default]

        if series_provider_timeout:
            sickrage.app.config.general.series_provider_timeout = try_int(series_provider_timeout)

        if time_preset:
            sickrage.app.config.gui.time_preset_w_seconds = time_preset
            sickrage.app.config.gui.time_preset = sickrage.app.config.gui.time_preset_w_seconds.replace(":%S", "")

        sickrage.app.config.gui.timezone_display = TimezoneDisplay[timezone_display]

        sickrage.app.config.general.api_v1_key = api_key

        sickrage.app.config.general.enable_https = checkbox_to_value(enable_https)

        if not change_https_cert(https_cert):
            results += ["Unable to create directory " + os.path.normpath(https_cert) + ", https cert directory not changed."]

        if not change_https_key(https_key):
            results += ["Unable to create directory " + os.path.normpath(https_key) + ", https key directory not changed."]

        sickrage.app.config.general.handle_reverse_proxy = checkbox_to_value(handle_reverse_proxy)

        sickrage.app.config.gui.theme_name = UITheme[theme_name]

        sickrage.app.config.general.default_page = DefaultHomePage[default_page]

        sickrage.app.config.general.max_queue_workers = try_int(max_queue_workers)

        sickrage.app.config.save()

        if auth_method_changed:
            return self.redirect('/logout')

        if len(results) > 0:
            [sickrage.app.log.error(x) for x in results]
            sickrage.app.alerts.error(_('Error(s) Saving Configuration'), '<br>\n'.join(results))
        else:
            sickrage.app.alerts.message(_('[GENERAL] Configuration Saved to Database'))

        return self.redirect("/config/general/")



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import os

from guessit import guessit

import sickrage

MOVIE_ATTRIBUTES = {
    'title': 'title',
    'year': 'year',
    'source': 'source',
    'release_group': 'release_group',
    'resolution': 'screen_size',
    'video_codec': 'video_codec',
    'audio_codec': 'audio_codec',
}

EPISODE_ATTRIBUTES = {
    'series': 'title',
    'season': 'season',
    'episode': 'episode',
    'title': 'episode_title',
    'year': 'year',
    'source': 'source',
    'release_group': 'release_group',
    'resolution': 'screen_size',
    'video_codec': 'video_codec',
    'audio_codec': 'audio_codec',
}


def refine(video, release_name=None, release_file=None, extension='release', **kwargs):
    """Refine a video by using the original release name.

    The refiner will first try:
    - Read the file video_name.<extension> seeking for a release name
    - If no release name, it will read the release_file seeking for a release name
    - If no release name, it will use the release_name passed as an argument
    - If no release name, then no change in the video object is made

    When a release name is found, the video object will be enhanced using the guessit properties extracted from it.

    Several :class:`~subliminal.video.Video` attributes can be found:

      * :attr:`~subliminal.video.Video.title`
      * :attr:`~subliminal.video.Video.series`
      * :attr:`~subliminal.video.Video.season`
      * :attr:`~subliminal.video.Video.episode`
      * :attr:`~subliminal.video.Video.year`
      * :attr:`~subliminal.video.Video.source`
      * :attr:`~subliminal.video.Video.release_group`
      * :attr:`~subliminal.video.Video.resolution`
      * :attr:`~subliminal.video.Video.video_codec`
      * :attr:`~subliminal.video.Video.audio_codec`

    :param video: the video to refine.
    :type video: subliminal.video.Video
    :param str release_name: the release name to be used.
    :param str release_file: the release file to be used
    :param str extension: the release file extension.
    """
    sickrage.app.log.debug(f'Starting release refiner [extension={extension}, release_name={release_name}, release_file={release_file}]')
    dirpath, filename = os.path.split(video.name)
    dirpath = dirpath or '.'
    fileroot, fileext = os.path.splitext(filename)
    release_file = get_release_file(dirpath, fileroot, extension) or release_file
    release_name = get_release_name(release_file) or release_name

    if not release_name:
        sickrage.app.log.debug(f'No release name for {video.name!r}')
        return

    release_path = os.path.join(dirpath, release_name + fileext)
    sickrage.app.log.debug('Guessing using {path}', {'path': release_path})

    guess = guessit(release_path)
    attributes = MOVIE_ATTRIBUTES if guess.get('type') == 'movie' else EPISODE_ATTRIBUTES
    for key, value in attributes.items():
        old_value = getattr(video, key)
        new_value = guess.get(value)

        if new_value and old_value != new_value:
            setattr(video, key, new_value)
            sickrage.app.log.debug(f'Attribute {key} changed from {old_value!r} to {new_value!r}')


def get_release_file(dirpath, filename, extension):
    """Return the release file that should contain the release name for a given a `dirpath`, `filename` and `extension`.

    :param dirpath: the file base folder
    :type dirpath: str
    :param filename: the file name without extension
    :type filename: str
    :param extension:
    :type extension: the file extension
    :return: the release file if the file exists
    :rtype: str
    """
    release_file = os.path.join(dirpath, filename + '.' + extension)

    # skip if info file doesn't exist
    if os.path.isfile(release_file):
        sickrage.app.log.debug(f'Found release file {release_file}')
        return release_file


def get_release_name(release_file):
    """Given a `release_file` it will return the release name.

    :param release_file: the text file that contains the release name
    :type release_file: str
    :return: the release name
    :rtype: str
    """
    if not release_file or not os.path.isfile(release_file):
        return

    with open(release_file, 'r') as f:
        release_name = f.read().strip()

    # skip if no release name was found
    if not release_name:
        sickrage.app.log.warning(f'Release file {release_file} does not contain a release name')

    return release_name



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

import os

from tornado.web import authenticated

import sickrage
from sickrage.core.helpers import backup_app_data, checkbox_to_value, restore_config_zip
from sickrage.core.webserver import ConfigWebHandler
from sickrage.core.webserver.handlers.base import BaseHandler


class ConfigBackupRestoreHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        return self.render('config/backup_restore.mako',
                           submenu=ConfigWebHandler.menu,
                           title=_('Config - Backup/Restore'),
                           header=_('Backup/Restore'),
                           topmenu='config',
                           controller='config',
                           action='backup_restore')


class ConfigBackupHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        backup_dir = self.get_argument('backupDir')

        final_result = ''

        if backup_dir:
            if backup_app_data(backup_dir):
                final_result += _("Backup SUCCESSFUL")
            else:
                final_result += _("Backup FAILED!")
        else:
            final_result += _("You need to choose a folder to save your backup to first!")

        final_result += "<br>\n"

        return final_result


class ConfigRestoreHandler(BaseHandler):
    @authenticated
    def get(self, *args, **kwargs):
        backup_file = self.get_argument('backupFile')
        restore_database = self.get_argument('restore_database')
        restore_config = self.get_argument('restore_config')
        restore_cache = self.get_argument('restore_cache')

        final_result = ''

        if backup_file:
            source = backup_file
            target_dir = os.path.join(sickrage.app.data_dir, 'restore')

            restore_database = checkbox_to_value(restore_database)
            restore_config = checkbox_to_value(restore_config)
            restore_cache = checkbox_to_value(restore_cache)

            if restore_config_zip(source, target_dir, restore_database, restore_config, restore_cache):
                final_result += _("Successfully extracted restore files to " + target_dir)
                final_result += _("<br>Restart sickrage to complete the restore.")
            else:
                final_result += _("Restore FAILED")
        else:
            final_result += _("You need to select a backup file to restore!")

        final_result += "<br>\n"

        return final_result


class SaveBackupRestoreHandler(BaseHandler):
    @authenticated
    def post(self, *args, **kwargs):
        return self.redirect("/config/backuprestore/")



from __future__ import absolute_import, division, print_function

from trakt.interfaces.base import authenticated
from trakt.interfaces.sync.core.mixins import Get, Delete


class SyncPlaybackInterface(Get, Delete):
    path = 'sync/playback'

    @authenticated
    def shows(self, store=None, **kwargs):
        raise NotImplementedError()

    @authenticated
    def episodes(self, store=None, **kwargs):
        return self.get(
            'episodes',
            store,
            **kwargs
        )



from __future__ import absolute_import, division, print_function

from trakt.core.helpers import from_iso8601_datetime, to_iso8601_datetime, deprecated
from trakt.objects.core.helpers import update_attributes
from trakt.objects.video import Video


class Episode(Video):
    def __init__(self, client, keys=None, index=None):
        super(Episode, self).__init__(client, keys, index)

        self.show = None
        """
        :type: :class:`trakt.objects.show.Show`

        Show
        """

        self.season = None
        """
        :type: :class:`trakt.objects.season.Season`

        Season
        """

        self.title = None
        """
        :type: :class:`~python:str`

        Title
        """

        self.first_aired = None
        """
        :type: :class:`~python:datetime.datetime`

        First air date
        """

        self.updated_at = None
        """
        :type: :class:`~python:datetime.datetime`

        Updated date/time
        """

        self.available_translations = None
        """
        :type: :class:`~python:list`

        Available translations (for title, overview, etc..)
        """

    def to_identifier(self):
        """Retrieve the episode identifier.

        :return: Episode identifier/definition
        :rtype: :class:`~python:dict`
        """

        __, number = self.pk

        return {
            'number': number
        }

    @deprecated('Episode.to_info() has been moved to Episode.to_dict()')
    def to_info(self):
        """**Deprecated:** use the :code:`to_dict()` method instead."""
        return self.to_dict()

    def to_dict(self):
        """Dump episode to a dictionary.

        :return: Episode dictionary
        :rtype: :class:`~python:dict`
        """

        result = self.to_identifier()

        result.update({
            'title': self.title,

            'watched': 1 if self.is_watched else 0,
            'collected': 1 if self.is_collected else 0,

            'plays': self.plays if self.plays is not None else 0,
            'in_watchlist': self.in_watchlist if self.in_watchlist is not None else 0,
            'progress': self.progress,

            'last_watched_at': to_iso8601_datetime(self.last_watched_at),
            'collected_at': to_iso8601_datetime(self.collected_at),
            'paused_at': to_iso8601_datetime(self.paused_at),

            'ids': dict([
                (key, value) for (key, value) in self.keys[1:]  # NOTE: keys[0] is the (<season>, <episode>) identifier
            ])
        })

        if self.rating:
            result['rating'] = self.rating.value
            result['rated_at'] = to_iso8601_datetime(self.rating.timestamp)

        # Extended Info
        if self.first_aired:
            result['first_aired'] = to_iso8601_datetime(self.first_aired)

        if self.updated_at:
            result['updated_at'] = to_iso8601_datetime(self.updated_at)

        if self.overview:
            result['overview'] = self.overview

        if self.available_translations:
            result['available_translations'] = self.available_translations

        return result

    def _update(self, info=None, **kwargs):
        if not info:
            return

        super(Episode, self)._update(info, **kwargs)

        update_attributes(self, info, [
            'title',

            # Extended Info
            'available_translations'
        ])

        # Extended Info
        if 'first_aired' in info:
            self.first_aired = from_iso8601_datetime(info.get('first_aired'))

        if 'updated_at' in info:
            self.updated_at = from_iso8601_datetime(info.get('updated_at'))

    @classmethod
    def _construct(cls, client, keys, info=None, index=None, **kwargs):
        episode = cls(client, keys, index=index)
        episode._update(info, **kwargs)

        return episode

    def __repr__(self):
        if self.show and self.title:
            return '<Episode %r - S%02dE%02d - %r>' % (self.show.title, self.pk[0], self.pk[1], self.title)

        if self.show:
            return '<Episode %r - S%02dE%02d>' % (self.show.title, self.pk[0], self.pk[1])

        if self.title:
            return '<Episode S%02dE%02d - %r>' % (self.pk[0], self.pk[1], self.title)

        return '<Episode S%02dE%02d>' % self.pk



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################

from marshmallow import Schema, fields


class BaseSchema(Schema):
    class Meta:
        ordered = True


class NotAuthorizedSchema(BaseSchema):
    error = fields.String(
        required=False,
        description="Authorization error",
        default="Not Authorized",
    )


class NotFoundSchema(BaseSchema):
    error = fields.String(
        required=False,
        description="Not Found error",
        default="Not Found",
    )


class BaseSuccessSchema(BaseSchema):
    success = fields.Boolean(
        required=True,
        description='This is always "True" when a request succeeds',
        example=True,
    )


class BaseErrorSchema(BaseSchema):
    success = fields.Boolean(
        required=True,
        description='This is always "False" when a request fails',
        example=False,
    )


class BadRequestSchema(BaseErrorSchema):
    errors = fields.Dict(
        required=False,
        description="Attached request validation errors",
        example={"name": ["Missing data for required field."]},
    )



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SiCKRAGE.
#
# SiCKRAGE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SiCKRAGE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import traceback

from sqlalchemy import orm

import sickrage
from sickrage.core.databases.main import MainDB
from sickrage.core.exceptions import EpisodeNotFoundException
from sickrage.core.tv.show.helpers import find_show
from sickrage.core.websession import WebSession


def get_scene_numbering(series_id, series_provider_id, season, episode, fallback_to_xem=True):
    """
    Returns a tuple, (season, episode), with the scene numbering (if there is one),
    otherwise returns the xem numbering (if fallback_to_xem is set), otherwise
    returns the TVDB numbering.
    (so the return values will always be set)

    :param series_id: int
    :param season: int
    :param episode: int
    :param fallback_to_xem: bool If set (the default), check xem for matches if there is no local scene numbering
    :return: (int, int) a tuple with (season, episode)
    """
    show_obj = find_show(series_id, series_provider_id)
    if not show_obj:
        return -1, -1

    result = find_scene_numbering(series_id, series_provider_id, season, episode)
    if result:
        return result

    if fallback_to_xem:
        xem_result = find_xem_numbering(series_id, series_provider_id, season, episode)
        if xem_result:
            return xem_result

    return -1, -1


def get_scene_absolute_numbering(series_id, series_provider_id, absolute_number, fallback_to_xem=True):
    """
    Returns absolute number, with the scene numbering (if there is one),
    otherwise returns the xem numbering (if fallback_to_xem is set), otherwise
    returns the TVDB numbering.
    (so the return values will always be set)

    :param series_id: int
    ;param absolute_number: int
    :param fallback_to_xem: bool If set (the default), check xem for matches if there is no local scene numbering
    :return: int absolute number
    """
    show_obj = find_show(series_id, series_provider_id)
    if not show_obj:
        return -1

    result = find_scene_absolute_numbering(series_id, series_provider_id, absolute_number)
    if result:
        return result

    if fallback_to_xem:
        xem_result = find_xem_absolute_numbering(series_id, series_provider_id, absolute_number)
        if xem_result:
            return xem_result

    return -1


def get_series_provider_numbering(series_id, series_provider_id, scene_season, scene_episode, fallback_to_xem=True):
    """
    Returns a tuple, (season, episode) with the TVDB numbering for (sceneSeason, sceneEpisode)
    (this works like the reverse of get_scene_numbering)
    """
    session = sickrage.app.main_db.session()

    try:
        dbData = session.query(MainDB.TVEpisode).filter_by(
            series_id=series_id,
            series_provider_id=series_provider_id,
            scene_season=scene_season,
            scene_episode=scene_episode
        ).one()

        return dbData.season, dbData.episode
    except orm.exc.NoResultFound:
        if fallback_to_xem:
            return get_series_provider_numbering_from_xem_numbering(series_id, series_provider_id, scene_season, scene_episode)
        return -1, -1


def get_series_provider_absolute_numbering(series_id, series_provider_id, scene_absolute_number, fallback_to_xem=True, scene_season=None):
    """
    Returns a tuple, (season, episode, absolute_number) with the TVDB absolute numbering for (sceneAbsoluteNumber)
    (this works like the reverse of get_absolute_numbering)
    """
    session = sickrage.app.main_db.session()

    try:
        if scene_season is None:
            dbData = session.query(MainDB.TVEpisode).filter_by(
                series_id=series_id,
                series_provider_id=series_provider_id,
                scene_absolute_number=scene_absolute_number
            ).one()
        else:
            dbData = session.query(MainDB.TVEpisode).filter_by(
                series_id=series_id,
                series_provider_id=series_provider_id,
                scene_absolute_number=scene_absolute_number,
                scene_season=scene_season
            ).one()
        return dbData.absolute_number
    except (orm.exc.MultipleResultsFound, orm.exc.NoResultFound):
        if fallback_to_xem:
            return get_series_provider_absolute_numbering_from_xem_numbering(series_id, series_provider_id, scene_absolute_number, scene_season)

        return -1


def get_series_provider_numbering_from_xem_numbering(series_id, series_provider_id, xem_season, xem_episode):
    """
    Reverse of find_xem_numbering: lookup series_provider_id season and episode using xem numbering

    :param series_id: int
    :param xem_season: int
    :param xem_episode: int
    :return: (int, int) a tuple of (season, episode)
    """
    session = sickrage.app.main_db.session()

    xem_refresh(series_id, series_provider_id)

    try:
        dbData = session.query(MainDB.TVEpisode).filter_by(series_id=series_id, series_provider_id=series_provider_id, xem_season=xem_season,
                                                           xem_episode=xem_episode).one()
        return dbData.season, dbData.episode
    except (orm.exc.NoResultFound, orm.exc.MultipleResultsFound):
        return -1, -1


def get_series_provider_absolute_numbering_from_xem_numbering(series_id, series_provider_id, xem_absolute_number, xem_season=None):
    """
    Reverse of find_xem_numbering: lookup series_provider_id season and episode using xem numbering

    :param series_id: int
    :param xem_absolute_number: int
    :return: int
    """
    session = sickrage.app.main_db.session()

    xem_refresh(series_id, series_provider_id)

    try:
        if xem_season is None:
            dbData = session.query(MainDB.TVEpisode).filter_by(
                series_id=series_id,
                series_provider_id=series_provider_id,
                xem_absolute_number=xem_absolute_number).one()
        else:
            dbData = session.query(MainDB.TVEpisode).filter_by(
                series_id=series_id,
                series_provider_id=series_provider_id,
                xem_absolute_number=xem_absolute_number,
                xem_season=xem_season).one()
        return dbData.absolute_number
    except (orm.exc.MultipleResultsFound, orm.exc.NoResultFound):
        return -1


def get_scene_numbering_for_show(series_id, series_provider_id):
    """
    Returns a dict of (season, episode) : (sceneSeason, sceneEpisode) mappings
    for an entire show.  Both the keys and values of the dict are tuples.
    Will be empty if there are no scene numbers set
    """
    session = sickrage.app.main_db.session()

    result = {}

    for dbData in session.query(MainDB.TVEpisode).filter_by(series_id=series_id):
        if dbData.series_provider_id != series_provider_id or (dbData.scene_season or dbData.scene_episode) == -1:
            continue

        season = dbData.season
        episode = dbData.episode
        scene_season = dbData.scene_season
        scene_episode = dbData.scene_episode

        result[(season, episode)] = (scene_season, scene_episode)

    return result


def get_scene_absolute_numbering_for_show(series_id, series_provider_id):
    """
    Returns a dict of (season, episode) : (sceneSeason, sceneEpisode) mappings
    for an entire show.  Both the keys and values of the dict are tuples.
    Will be empty if there are no scene numbers set
    """
    session = sickrage.app.main_db.session()

    result = {}

    for dbData in session.query(MainDB.TVEpisode).filter_by(series_id=series_id):
        if dbData.series_provider_id != series_provider_id or dbData.scene_absolute_number == -1:
            continue

        absolute_number = dbData.absolute_number
        scene_absolute_number = dbData.scene_absolute_number

        result[absolute_number] = scene_absolute_number

    return result


def get_xem_numbering_for_show(series_id, series_provider_id):
    """
    Returns a dict of (season, episode) : (sceneSeason, sceneEpisode) mappings
    for an entire show.  Both the keys and values of the dict are tuples.
    Will be empty if there are no scene numbers set in xem
    """
    session = sickrage.app.main_db.session()

    xem_refresh(series_id, series_provider_id)

    result = {}

    for dbData in session.query(MainDB.TVEpisode).filter_by(series_id=series_id):
        if dbData.series_provider_id != series_provider_id or (dbData.xem_season or dbData.xem_episode) == -1:
            continue

        season = dbData.season
        episode = dbData.episode
        xem_season = dbData.xem_season
        xem_episode = dbData.xem_episode

        result[(season, episode)] = (xem_season, xem_episode)

    return result


def get_xem_absolute_numbering_for_show(series_id, series_provider_id):
    """
    Returns a dict of (season, episode) : (sceneSeason, sceneEpisode) mappings
    for an entire show.  Both the keys and values of the dict are tuples.
    Will be empty if there are no scene numbers set in xem
    """
    session = sickrage.app.main_db.session()

    xem_refresh(series_id, series_provider_id)

    result = {}

    for dbData in session.query(MainDB.TVEpisode).filter_by(series_id=series_id):
        if dbData.series_provider_id != series_provider_id or dbData.xem_absolute_number == -1:
            continue

        absolute_number = dbData.absolute_number
        xem_absolute_number = dbData.xem_absolute_number

        result[absolute_number] = xem_absolute_number

    return result


def find_scene_numbering(series_id, series_provider_id, season, episode):
    """
    Same as get_scene_numbering(), but returns None if scene numbering is not set
    """
    session = sickrage.app.main_db.session()

    try:
        dbData = session.query(MainDB.TVEpisode).filter_by(
            series_id=series_id,
            series_provider_id=series_provider_id,
            season=season,
            episode=episode
        ).filter(MainDB.TVEpisode.scene_season != -1 and MainDB.TVEpisode.scene_episode != -1).one()
        return dbData.scene_season, dbData.scene_episode
    except orm.exc.NoResultFound:
        return


def find_scene_absolute_numbering(series_id, series_provider_id, absolute_number):
    """
    Same as get_scene_numbering(), but returns None if scene numbering is not set
    """
    session = sickrage.app.main_db.session()

    try:
        dbData = session.query(MainDB.TVEpisode).filter_by(
            series_id=series_id,
            series_provider_id=series_provider_id,
            absolute_number=absolute_number
        ).filter(MainDB.TVEpisode.scene_absolute_number != -1).one()

        return dbData.scene_absolute_number
    except (orm.exc.MultipleResultsFound, orm.exc.NoResultFound):
        return


def find_xem_numbering(series_id, series_provider_id, season, episode):
    """
    Returns the scene numbering, as retrieved from xem.
    Refreshes/Loads as needed.

    :param series_id: int
    :param season: int
    :param episode: int
    :return: (int, int) a tuple of scene_season, scene_episode, or None if there is no special mapping.
    """
    session = sickrage.app.main_db.session()

    xem_refresh(series_id, series_provider_id)

    try:
        dbData = session.query(MainDB.TVEpisode).filter_by(
            series_id=series_id,
            series_provider_id=series_provider_id,
            season=season,
            episode=episode
        ).filter(MainDB.TVEpisode.xem_season != -1, MainDB.TVEpisode.xem_episode != -1).one()

        return dbData.xem_season, dbData.xem_episode
    except orm.exc.NoResultFound:
        return


def find_xem_absolute_numbering(series_id, series_provider_id, absolute_number):
    """
    Returns the scene numbering, as retrieved from xem.
    Refreshes/Loads as needed.

    :param series_id: int
    :param absolute_number: int
    :return: int
    """
    session = sickrage.app.main_db.session()

    xem_refresh(series_id, series_provider_id)

    try:
        dbData = session.query(MainDB.TVEpisode).filter_by(series_id=series_id, series_provider_id=series_provider_id,
                                                           absolute_number=absolute_number).filter(MainDB.TVEpisode.xem_absolute_number != -1).one()
        return dbData.xem_absolute_number
    except (orm.exc.NoResultFound, orm.exc.MultipleResultsFound):
        return


def set_scene_numbering(series_id, series_provider_id, season=None, episode=None, absolute_number=None, scene_season=None, scene_episode=None,
                        scene_absolute=None):
    """
    Set scene numbering for a season/episode or absolute.
    To clear the scene numbering, leave both scene_season and scene_episode or scene_absolute as None.
    """
    session = sickrage.app.main_db.session()

    if season and episode:
        if scene_season is not None and scene_episode is not None:
            if session.query(MainDB.TVEpisode).filter_by(series_id=series_id, series_provider_id=series_provider_id, scene_season=scene_season,
                                                         scene_episode=scene_episode).count():
                return False

        dbData = session.query(MainDB.TVEpisode).filter_by(series_id=series_id, series_provider_id=series_provider_id, season=season, episode=episode).one()
        dbData.scene_season = scene_season if scene_season is not None else -1
        dbData.scene_episode = scene_episode if scene_episode is not None else -1
    elif absolute_number:
        if scene_absolute is not None:
            if session.query(MainDB.TVEpisode).filter_by(series_id=series_id, series_provider_id=series_provider_id,
                                                         scene_absolute_number=scene_absolute).count():
                return False

        dbData = session.query(MainDB.TVEpisode).filter_by(series_id=series_id, series_provider_id=series_provider_id, absolute_number=absolute_number).one()
        dbData.scene_absolute_number = scene_absolute if scene_absolute is not None else -1

    session.commit()

    return True


def xem_refresh(series_id, series_provider_id, force=False):
    """
    Refresh data from xem for a tv show

    :param series_id: int
    :param series_provider_id: int
    :param force: boolean
    """
    max_refresh_age_secs = 86400  # 1 day

    show_object = find_show(series_id, series_provider_id)
    if not show_object:
        return

    if datetime.datetime.now() > (show_object.last_xem_refresh + datetime.timedelta(seconds=max_refresh_age_secs)) or force:
        sickrage.app.log.debug('Looking up XEM scene mapping for show %s on %s' % (series_id, show_object.series_provider.name))

        # mark xem refreshed
        show_object.last_xem_refresh = datetime.datetime.now()
        show_object.save()

        try:
            try:
                # XEM MAP URL
                url = "http://thexem.de/map/havemap?origin=%s" % show_object.series_provider.xem_origin
                parsed_json = WebSession().get(url).json()
                if not parsed_json or 'data' not in parsed_json:
                    raise ValueError
            except ValueError:
                sickrage.app.log.warning("Resulting JSON from XEM isn't correct, not parsing it")
                return

            if series_id not in map(int, parsed_json['data']):
                sickrage.app.log.info('No XEM data for show {} on {}'.format(show_object.name, show_object.series_provider.name))
                # for episode_object in show_object.episodes:
                #     episode_object.xem_season = -1
                #     episode_object.xem_episode = -1
                #     episode_object.xem_absolute_number = -1
                #     episode_object.save()
                return

            try:
                # XEM API URL
                url = "http://thexem.de/map/all?id={}&origin={}&destination=scene".format(series_id, show_object.series_provider.xem_origin)
                parsed_json = WebSession().get(url).json()
                if not parsed_json or 'result' not in parsed_json or 'data' not in parsed_json:
                    raise ValueError
            except ValueError:
                sickrage.app.log.warning("Resulting JSON from XEM isn't correct, not parsing it")
                return

            if 'success' not in parsed_json['result']:
                sickrage.app.log.info('No XEM data for show {} on {}'.format(show_object.name, show_object.series_provider.name))
                return

            for entry in parsed_json['data']:
                try:
                    episode_object = show_object.get_episode(season=entry[show_object.series_provider.xem_origin]['season'],
                                                             episode=entry[show_object.series_provider.xem_origin]['episode'])
                except EpisodeNotFoundException:
                    continue

                if 'scene' in entry:
                    episode_object.xem_season = entry['scene']['season']
                    episode_object.xem_episode = entry['scene']['episode']
                    episode_object.xem_absolute_number = entry['scene']['absolute']

                if 'scene_2' in entry:  # for doubles
                    episode_object.xem_season = entry['scene_2']['season']
                    episode_object.xem_episode = entry['scene_2']['episode']
                    episode_object.xem_absolute_number = entry['scene_2']['absolute']

                episode_object.save()
        except Exception as e:
            sickrage.app.log.debug("Exception while refreshing XEM data for show {} on {}: {}".format(series_id, show_object.series_provider.name, e))
            sickrage.app.log.debug(traceback.format_exc())


def get_absolute_number_from_season_and_episode(series_id, series_provider_id, season, episode):
    """
    Find the absolute number for a show episode

    :param series_id: Series ID
    :param series_provider_id: Series Provider ID
    :param season: Season number
    :param episode: Episode number
    :return: The absolute number
    """
    session = sickrage.app.main_db.session()

    absolute_number = None

    show_object = find_show(series_id, series_provider_id)
    if not show_object:
        return

    if season and episode:
        try:
            dbData = session.query(MainDB.TVEpisode).filter_by(series_id=series_id, series_provider_id=series_provider_id, season=season,
                                                               episode=episode).one()
            absolute_number = dbData.absolute_number
            sickrage.app.log.debug("Found absolute number %s for show %s S%02dE%02d" % (absolute_number, show_object.name, season, episode))
        except orm.exc.NoResultFound:
            sickrage.app.log.debug("No entries for absolute number for show %s S%02dE%02d" % (show_object.name, season, episode))

    return absolute_number



# ##############################################################################
#  Author: echel0n <echel0n@sickrage.ca>
#  URL: https://sickrage.ca/
#  Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#  -
#  This file is part of SiCKRAGE.
#  -
#  SiCKRAGE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  -
#  SiCKRAGE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  -
#  You should have received a copy of the GNU General Public License
#  along with SiCKRAGE.  If not, see <http://www.gnu.org/licenses/>.
# ##############################################################################


import functools
import json
import re
import time
from base64 import urlsafe_b64decode
from datetime import datetime
from operator import itemgetter
from urllib.parse import urljoin, quote

import requests
from six import text_type

import sickrage
from sickrage.core.enums import SeriesProviderID
from sickrage.core.websession import WebSession
from sickrage.series_providers import SeriesProvider
from sickrage.series_providers.exceptions import SeriesProviderNotAuthorized

try:
    import gzip
except ImportError:
    gzip = None


def login_required(f):
    @functools.wraps(f)
    def wrapper(obj, *args, **kwargs):
        if not obj.jwt_token:
            obj.authenticate()

        for i in range(3):
            try:
                return f(obj, *args, **kwargs)
            except SeriesProviderNotAuthorized:
                obj.authenticate()

        sickrage.app.log.debug("Unable to authenticate to TheTVDB")

    return wrapper


def to_lowercase(iterable):
    if type(iterable) is dict:
        for key in iterable.copy().keys():
            iterable[key.lower()] = iterable.pop(key)
            if type(iterable[key.lower()]) is dict or type(iterable[key.lower()]) is list:
                iterable[key.lower()] = to_lowercase(iterable[key.lower()])
    elif type(iterable) is list:
        for i, item in enumerate(iterable.copy()):
            iterable[i] = to_lowercase(item)

    return iterable


class TheTVDBActors(list):
    """Holds all Actor instances for a show
    """
    pass


class TheTVDBActor(dict):
    """Represents a single actor. Should contain..

    id,
    image,
    name,
    role,
    sortorder
    """

    def __repr__(self):
        return "<Actor \"{}\">".format(self.get("name"))


class TheTVDB(SeriesProvider):
    """Create easy-to-use interface to name of season/episode name
    """

    def __init__(self):
        super(TheTVDB, self).__init__(SeriesProviderID.THETVDB)
        self.apikey = 'F9C450E78D99172E'
        self.trakt_id = 'tvdb'
        self.xem_origin = 'tvdb'
        self.icon = 'thetvdb16.png'
        self.show_url = 'http://thetvdb.com/?tab=series&id='
        self.dvd_order = False

        self.api = {
            'version': '3.0.0',
            'base': "https://api.thetvdb.com",
            'login': '/login',
            'refresh': '/refresh_token',
            'languages': '/languages',
            'languageInfo': '/languages/{id}',
            'getSeries': "/search/series?name={name}",
            'getSeriesIMDB': "/search/series?imdbId={id}",
            'getSeriesZap2It': "/search/series?zap2itId={id}",
            'series': "/series/{id}",
            'episodes': "/series/{id}/episodes",
            'episode_info': "/episodes/{id}",
            'actors': "/series/{id}/actors",
            'updated': "/updated/query?fromTime={time}",
            'images': {
                'fanart': "/series/{id}/images/query?keyType=fanart&subKey=graphical",
                'banner': "/series/{id}/images/query?keyType=fanart&subKey=text",
                'poster': "/series/{id}/images/query?keyType=poster",
                'series': "/series/{id}/images/query?keyType=series",
                'season': "/series/{id}/images/query?keyType=season&subKey={season}",
                'seasonwide': "/series/{id}/images/query?keyType=seasonwide&subKey={season}",
                'params': "/series/{id}/images/query/params",
                'prefix': "https://www.thetvdb.com/banners/{value}"
            }
        }

    @property
    def jwt_token(self):
        return getattr(self, '_jwt_token', None)

    @jwt_token.setter
    def jwt_token(self, value):
        if self.jwt_token != value:
            setattr(self, '_jwt_token', value)
            self.jwt_payload = self.get_jwt_payload(self.jwt_token)

    @property
    def jwt_payload(self):
        return getattr(self, '_jwt_payload', {})

    @jwt_payload.setter
    def jwt_payload(self, value):
        if self.jwt_payload != value:
            setattr(self, '_jwt_payload', value)

    @property
    def jwt_expiration(self):
        return self.jwt_payload.get('exp', time.time())

    @property
    def jwt_time_remaining(self):
        return max(self.jwt_expiration - time.time(), 0)

    @property
    def jwt_is_expired(self):
        return self.jwt_expiration <= time.time()

    def logout(self):
        self.jwt_token = None

    def _refresh(self):
        resp = self._request('get', self.api['refresh'])
        if resp and 'token' in resp:
            self.jwt_token = resp['token']

    def _login(self):
        resp = self._request('post', self.api['login'], json={'apikey': self.apikey})
        if resp and 'token' in resp:
            self.jwt_token = resp['token']

    def authenticate(self):
        try:
            if not self.jwt_token or self.jwt_is_expired:
                return self._login()
            elif self.jwt_time_remaining < 7200:
                return self._refresh()
        except Exception as e:
            self.logout()

    def jwt_decode(self, data):
        # make sure data is binary
        if isinstance(data, text_type):
            sickrage.app.log.debug('Encoding the JWT token as UTF-8')
            data = data.encode('utf-8')

        # pad the data to a multiple of 4 bytes
        remainder = len(data) % 4
        if remainder > 0:
            length = 4 - remainder
            sickrage.app.log.debug('Padding the JWT with {x} bytes'.format(x=length))
            data += b'=' * length

        # base64 decode the data
        data = urlsafe_b64decode(data)

        # convert the decoded json to a string
        data = data.decode('utf-8')

        # return the json string as a dict
        result = json.loads(data)
        sickrage.app.log.info('JWT Successfully decoded')
        return result

    def get_jwt_payload(self, token):
        result = {}

        try:
            __, payload, __ = token.split('.')
        except AttributeError:
            sickrage.app.log.debug('Unable to extract payload from JWT: {}'.format(token))
        else:
            result = self.jwt_decode(payload)
            sickrage.app.log.debug('Payload extracted from JWT: {}'.format(result))
        finally:
            return result

    def _request(self, method, url, language='en', **kwargs):
        self.headers.update({'Authorization': 'Bearer {}'.format(self.jwt_token)})
        self.headers.update({'Content-type': 'application/json'})
        self.headers.update({'Accept-Language': language})
        self.headers.update({'Accept': 'application/vnd.thetvdb.v{}'.format(self.api['version'])})

        if not self.health:
            return None

        if sickrage.app.config.general.proxy_setting and sickrage.app.config.general.proxy_series_providers:
            proxy = sickrage.app.config.general.proxy_setting

        resp = WebSession(cache=self.cache).request(
            method, urljoin(self.api['base'], url), headers=self.headers,
            timeout=sickrage.app.config.general.series_provider_timeout, verify=False, **kwargs
        )

        if resp and resp.content:
            try:
                data = resp.json()
            except ValueError:
                sickrage.app.log.debug("Unable to parse data from TheTVDB")
                return None

            return to_lowercase(data)

        if resp is not None:
            if resp.status_code == 401:
                raise SeriesProviderNotAuthorized(resp.text)
            elif resp.status_code == 504:
                sickrage.app.log.debug("Unable to connect to TheTVDB")
                return None

            if 'application/json' in resp.headers.get('content-type', ''):
                err_msg = resp.json().get('Error', resp.text)
                sickrage.app.log.debug("Unable to get data from TheTVDB, Code: {code} Error: {err_msg!r}".format(code=resp.status_code, err_msg=err_msg))
                return None

    def _clean_data(self, data):
        """Cleans up strings returned by TheTVDB.com

        Issues corrected:
        - Replaces &amp; with &
        - Trailing whitespace
        """

        return data.replace("&amp;", "&").strip() if isinstance(data, str) else data

    @login_required
    def search(self, series, language='en', enable_cache=True):
        """This searches TheTVDB.com for the series by name, imdbid, or zap2itid
        and returns the result list
        """

        search_result = None

        if isinstance(series, int):
            if series in self.cache:
                if enable_cache:
                    search_result = self.cache[series]
            if not search_result:
                search_result = self._get_show_data(series, language=language)
        elif not re.search(r'tt\d+', series):
            sickrage.app.log.debug("Searching for show by name: {}".format(series))
            resp = self._request('get', self.api['getSeries'].format(name=quote(series), language=language))
            if resp and 'data' in resp:
                search_result = resp['data']
        else:
            sickrage.app.log.debug("Searching for show by imdbId: {}".format(series))
            resp = self._request('get', self.api['getSeriesIMDB'].format(id=series), language=language)
            if resp and 'data' in resp:
                if len(resp['data']) == 1:
                    search_result = resp['data'][0]

        if not search_result:
            sickrage.app.log.debug(f'Series search for {series} returned zero results, cannot find series on TheTVDB')

        return search_result

    @login_required
    def _get_show_data(self, sid, language='en'):
        """Takes a series ID, gets the episodes URL and parses the TVDB
        """

        languages = ['en']
        if language and language not in languages:
            languages.append(language)

        series_info = {}

        # Parse show information
        sickrage.app.log.debug('[{}]: Getting all series data from TheTVDB'.format(sid))

        for _language in languages:
            resp = self._request('get', self.api['series'].format(id=sid), language=_language)
            if not resp or 'data' not in resp:
                sickrage.app.log.debug("[{}]: Unable to locate show on TheTVDB".format(sid))
                return None

            series_info.update((k, v) for k, v in resp['data'].items() if v != '')

        # get series data
        for k, v in series_info.items():
            if v is not None:
                if k in ['banner', 'fanart', 'poster']:
                    v = self.api['images']['prefix'].format(value=v)
                elif isinstance(v, list):
                    v = '|'.join(v)
                else:
                    v = self._clean_data(v)

            self.cache.add_show_data(sid, k, v)

        # Parse episode data
        sickrage.app.log.debug('[{}]: Getting episode data from TheTVDB'.format(sid))

        episode_info = {}
        page = pages = 1

        while True:
            if page > pages:
                break

            for _language in languages:
                resp = self._request('get', self.api['episodes'].format(id=sid), language=_language, params={'page': page})
                if not resp or 'links' not in resp or 'data' not in resp:
                    break

                pages = resp['links']['last']

                for result in resp['data']:
                    if result['id'] not in episode_info:
                        episode_info[result['id']] = {}

                    episode_info[result['id']].update((k, v) for k, v in result.items() if v != '')

            page += 1

        if not len(episode_info):
            sickrage.app.log.debug('Series {} episodes incomplete on TheTVDB'.format(sid))
            return None

        episode_incomplete = False
        for cur_ep_id, cur_ep in episode_info.items():
            use_dvd = False

            if self.dvd_order:
                sickrage.app.log.debug('Using DVD ordering.')
                use_dvd = all([cur_ep.get('dvdseason'), cur_ep.get('dvdepisodenumber')])

            seasnum, epno = cur_ep.get('airedseason'), cur_ep.get('airedepisodenumber')
            if use_dvd:
                seasnum, epno = cur_ep.get('dvdseason'), cur_ep.get('dvdepisodenumber')

            if seasnum is None or epno is None:
                episode_incomplete = True
                continue

            seas_no = int(float(seasnum))
            ep_no = int(float(epno))

            for k, v in cur_ep.items():
                k = k.lower()
                if v is not None:
                    if isinstance(v, list):
                        v = '|'.join(v)
                    else:
                        v = self._clean_data(v)

                self.cache.add_item(sid, seas_no, ep_no, k, v)

            # add episode image url
            image_url = self.api['images']['prefix'].format(value='episodes/{}/{}.jpg'.format(sid, cur_ep['id']))
            self.cache.add_item(sid, seas_no, ep_no, 'filename', image_url)

        if episode_incomplete:
            sickrage.app.log.debug("{}: Series has incomplete season/episode numbers".format(sid))

        # set last updated
        self.cache.add_show_data(sid, 'last_updated', int(time.mktime(datetime.now().timetuple())))

        return self.cache[int(sid)]

    @login_required
    def image_key_types(self, sid, season=None, language='en'):
        key_types = {}

        resp = self._request('get', self.api['images']['params'].format(id=sid), language=language)
        if not resp or 'data' not in resp:
            return key_types

        for item in resp['data']:
            key_type = item['keytype']
            resolution = item['resolution']
            subkey = item['subkey']

            has_image = False
            if season and "season" in key_type:
                if season in subkey:
                    has_image = True
            elif key_type in ['fanart', 'series', 'poster']:
                if 'graphical' in subkey or len(resolution):
                    has_image = True

            if language not in key_types:
                key_types.update({
                    key_type: has_image
                })

        return key_types

    @login_required
    def images(self, sid, key_type='poster', season=None, language='en'):
        sickrage.app.log.debug('Getting {} images for {}'.format(key_type, sid))

        images = []

        languages = ['en']
        if language and language not in languages:
            languages.append(language)

        for _language in languages:
            key_types = self.image_key_types(sid, season, _language)
            if not key_types or key_type not in key_types:
                continue

            if not season:
                resp = self._request('get', self.api['images'][key_type].format(id=sid), language=_language)
                if resp and 'data' in resp:
                    images = resp['data']
            else:
                resp = self._request('get', self.api['images'][key_type].format(id=sid, season=season), language=_language)
                if resp and 'data' in resp:
                    images = resp['data']

        # unable to retrieve images in languages wanted
        if not images:
            return []

        for i, image in enumerate(images):
            if season and int(image['subkey']) != season:
                continue
            image["score"] = image["ratingsinfo"]["average"] * image["ratingsinfo"]["count"]
            for k, v in image.items():
                if not all([k, v]):
                    continue
                v = (v, self.api['images']['prefix'].format(value=v))[k in ['filename', 'thumbnail']]
                images[i][k] = v

        return [item for item in sorted(images, key=itemgetter("score"), reverse=True)]

    @login_required
    def actors(self, sid):
        sickrage.app.log.debug("Getting actors for {}".format(sid))

        actors = TheTVDBActors()

        resp = self._request('get', self.api['actors'].format(id=sid))
        if not resp or 'data' not in resp:
            sickrage.app.log.debug('Actors result returned zero')
            return actors

        for cur_actor in resp['data']:
            actor = TheTVDBActor()
            for k, v in cur_actor.items():
                if not all([k, v]):
                    continue
                v = (v, self.api['images']['prefix'].format(value=v))[k == 'image']
                actor[k] = v

            actors.append(cur_actor)

        return actors

    @login_required
    def updated(self, fromTime):
        resp = self._request('get', self.api['updated'].format(time=fromTime))
        if resp and 'data' in resp:
            return resp['data']

    @login_required
    def languages(self):
        resp = self._request('get', self.api['languages'])
        if resp and 'data' in resp:
            return sorted(resp['data'], key=lambda i: i['englishname'])

    @login_required
    def language_info(self, lang_id):
        resp = self._request('get', self.api['languageInfo'].format(id=lang_id))
        if resp and 'data' in resp:
            return resp['data']

    @property
    def health(self):
        for i in range(3):
            try:
                health = requests.get(urljoin(self.api['base'], 'health'), verify=False, timeout=30).ok
            except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout):
                pass
            else:
                break
        else:
            health = False

        if not health:
            sickrage.app.log.debug("TheTVDB API server is currently unreachable")
            return False

        return True

    def __repr__(self):
        return repr(self.cache)



"""Initial migration

Revision ID: 15
Revises:
Create Date: 2017-12-29 14:39:27.854291

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '15'
down_revision = '14'


def upgrade():
    conn = op.get_bind()
    meta = sa.MetaData(bind=conn)
    tv_episodes = sa.Table('tv_episodes', meta, autoload=True)

    with op.batch_alter_table("tv_episodes") as batch_op:
        if not hasattr(tv_episodes.c, 'xem_season'):
            batch_op.add_column(sa.Column('xem_season', sa.Integer, default=-1))
        if not hasattr(tv_episodes.c, 'xem_episode'):
            batch_op.add_column(sa.Column('xem_episode', sa.Integer, default=-1))
        if not hasattr(tv_episodes.c, 'xem_absolute_number'):
            batch_op.add_column(sa.Column('xem_absolute_number', sa.Integer, default=-1))


def downgrade():
    pass



class APIError(Exception):
    """
    API Error
    """

    def __init__(self, status, message, response):
        self._status = status
        self._message = message
        self._response = response

    @property
    def status(self):
        return self._status

    @property
    def message(self):
        return self._message

    @property
    def response(self):
        return self._response

    def __unicode__(self):
        return self.__class__.__name__ + ': ' + self.message

    def __repr__(self):
        return self.__unicode__()


class APIResourceDoesNotExist(APIError):
    """Custom exception when resource is not found."""
    pass


class APIUnauthorized(APIError):
    """Need JWT Token"""
    pass


class APITokenExpired(APIError):
    """JWT Token Expired"""
    pass
